from .std_msgs import ROSmsg
from .std_msgs import Header
from .std_msgs import Time
from .actionlib_msgs import GoalID
from .actionlib_msgs import GoalStatus
from .trajectory_msgs import JointTrajectoryPoint
from .trajectory_msgs import JointTrajectory


class JointTolerance(ROSmsg):
    """https://docs.ros.org/api/control_msgs/html/msg/JointTolerance.html
    """

    def __init__(self, name="", position=0., velocity=0., acceleration=0.):
        self.name = name
        self.position = position          # in radians or meters (for a revolute or prismatic joint, respectively)
        self.velocity = velocity          # in rad/sec or m/sec
        self.acceleration = acceleration  # in rad/sec^2 or m/sec^2


class FollowJointTrajectoryGoal(ROSmsg):
    """https://docs.ros.org/api/control_msgs/html/action/FollowJointTrajectory.html
    """

    def __init__(self, trajectory=None, path_tolerance=None,
                 goal_tolerance=None, goal_time_tolerance=None):
        self.trajectory = trajectory or JointTrajectory()  # trajectory_msgs/JointTrajectory
        self.path_tolerance = path_tolerance or []         # control_msgs/JointTolerance[]
        self.goal_tolerance = goal_tolerance or []         # control_msgs/JointTolerance[]
        self.goal_time_tolerance = goal_time_tolerance or Time(secs=1.)


class FollowJointTrajectoryActionGoal(ROSmsg):
    """https://docs.ros.org/fuerte/api/control_msgs/html/msg/FollowJointTrajectoryActionGoal.html
    """

    def __init__(self, header=None, goal_id=None, goal=None):
        self.header = header or Header()
        self.goal_id = goal_id or GoalID()  # actionlib_msgs/GoalID goal_id
        self.goal = goal or FollowJointTrajectoryGoal()  # FollowJointTrajectoryGoal goal


class FollowJointTrajectoryFeedback(ROSmsg):
    """https://docs.ros.org/fuerte/api/control_msgs/html/msg/FollowJointTrajectoryFeedback.html
    """

    def __init__(self, header=None, joint_names=None, desired=None, actual=None,
                 error=None):
        self.header = header or Header()
        self.joint_names = joint_names or []
        self.desired = desired or JointTrajectoryPoint()
        self.actual = actual or JointTrajectoryPoint()
        self.error = error or JointTrajectoryPoint()


class FollowJointTrajectoryActionFeedback(ROSmsg):
    """https://docs.ros.org/fuerte/api/control_msgs/html/msg/FollowJointTrajectoryActionFeedback.html
    """

    def __init__(self, header=None, status=None, feedback=None):
        self.header = header or Header()
        self.status = status or GoalStatus()
        self.feedback = feedback or FollowJointTrajectoryFeedback()


class FollowJointTrajectoryResult(ROSmsg):
    """https://docs.ros.org/fuerte/api/control_msgs/html/msg/FollowJointTrajectoryResult.html
    """

    SUCCESSFUL = 0
    INVALID_GOAL = -1
    INVALID_JOINTS = -2
    OLD_HEADER_TIMESTAMP = -3
    PATH_TOLERANCE_VIOLATED = -4
    GOAL_TOLERANCE_VIOLATED = -5

    def __init__(self, error_code=0, error_string=""):
        self.error_code = error_code
        self.error_string = error_string

    @classmethod
    def from_msg(cls, msg):
        error_code = msg['error_code']
        return cls(error_code)

    @property
    def human_readable(self):
        cls = type(self)
        for k, v in cls.__dict__.items():
            if v == self.error_code:
                return k
        return ''


class FollowJointTrajectoryActionResult(ROSmsg):
    """https://docs.ros.org/fuerte/api/control_msgs/html/msg/FollowJointTrajectoryActionResult.html
    """

    def __init__(self, header=None, status=None, result=None):
        self.header = header or Header()
        self.status = status or GoalStatus()
        self.result = result or FollowJointTrajectoryResult()

    @classmethod
    def from_msg(cls, msg):
        header = Header.from_msg(msg['header'])
        status = GoalStatus.from_msg(msg['status'])
        result = FollowJointTrajectoryResult.from_msg(msg['result'])
        return cls(header, status, result)



"""
Calculate a cartesian motion path (linear in tool space).

COMPAS FAB v0.18.2
"""
from ghpythonlib.componentbase import executingcomponent as component
from scriptcontext import sticky as st

from compas_fab.ghpython.components import coerce_frame
from compas_fab.ghpython.components import create_id


class PlanCartesianMotion(component):
    def RunScript(self, robot, planes, start_configuration, group, attached_collision_meshes, path_constraints, max_step, compute):

        key = create_id(self, 'trajectory')

        max_step = float(max_step) if max_step else 0.01
        path_constraints = list(path_constraints) if path_constraints else None
        attached_collision_meshes = list(attached_collision_meshes) if attached_collision_meshes else None

        if robot and robot.client and robot.client.is_connected and start_configuration and planes and compute:
            frames = [coerce_frame(plane) for plane in planes]
            st[key] = robot.plan_cartesian_motion(frames,
                                                  start_configuration=start_configuration,
                                                  group=group,
                                                  options=dict(
                                                      max_step=max_step,
                                                      path_constraints=path_constraints,
                                                      attached_collision_meshes=attached_collision_meshes
                                                  ))

        trajectory = st.get(key, None)
        return trajectory



from __future__ import print_function

import math

from roslibpy import Message
from roslibpy import Topic
from roslibpy.actionlib import Goal
from roslibpy.event_emitter import EventEmitterMixin

from compas_fab.backends.ros.messages.direct_ur import URGoal
from compas_fab.backends.ros.messages.direct_ur import URMovel
from compas_fab.backends.ros.messages.direct_ur import URPose
from compas_fab.backends.ros.messages.direct_ur import URPoseTrajectoryPoint

__all__ = ['DirectUrActionClient']


class DirectUrActionClient(EventEmitterMixin):
    """Direct UR Script action client to simulate an action interface
    on arbitrary URScript commands.
    """
    def __init__(self, ros, timeout=None,
                 omit_feedback=False, omit_status=False, omit_result=False):
        super(DirectUrActionClient, self).__init__()

        self.server_name = '/ur_driver/URScript'
        self.ros = ros
        self.timeout = timeout
        self.omit_feedback = omit_feedback
        self.omit_status = omit_status
        self.omit_result = omit_result
        self.goal = None

        self._received_status = False

        # Advertise the UR Script topic
        self._urscript_topic = Topic(ros, self.server_name, 'std_msgs/String')
        self._urscript_topic.advertise()

        # Create the topics associated with actionlib
        self.status_listener = Topic(ros, '/tool_velocity', 'geometry_msgs/TwistStamped')

        # Subscribe to the status topic
        if not self.omit_status:
            self.status_listener.subscribe(self._on_status_message)

        # If timeout specified, emit a 'timeout' event if the action server does not respond
        if self.timeout:
            self.ros.call_later(self.timeout / 1000., self._trigger_timeout)

        self._internal_state = 'idle'

    def _on_status_message(self, message):
        self._received_status = True

        self.goal.emit('status', message)

        twist = message['twist']
        total_velocity = math.fsum(list(twist['linear'].values()) + list(twist['angular'].values()))
        in_motion = total_velocity != 0.0

        if self._internal_state == 'idle':
            if in_motion:
                self._internal_state = 'executing_goal'
        elif self._internal_state == 'executing_goal':
            if not in_motion:
                self._internal_state = 'stopped'
                self.goal.emit('result', message)

    def _trigger_timeout(self):
        if not self._received_status:
            self.emit('timeout')

    def add_goal(self, goal):
        """Add a goal to this action client.

        Args:
            goal (:class:`.Goal`): Goal to add.
        """
        self.goal = goal

    def cancel(self):
        """Cancel all goals associated with this action client."""
        pass
        # self.cancel_topic.publish(Message())

    def dispose(self):
        """Unsubscribe and unadvertise all topics associated with this action client."""
        # And the UR Script topic
        self._urscript_topic.unadvertise()

        if not self.omit_status:
            self.status_listener.unsubscribe(self._on_status_message)

    def send_goal(self, goal, result_callback=None, timeout=None):
        """Send goal to the action server.

        Args:
            goal (:class:`URGoal`): The goal containing the script lines
            timeout (:obj:`int`): Timeout for the goal's result expressed in milliseconds.
            callback (:obj:`callable`): Function to be called when a result is received. It is a shorthand for hooking on the ``result`` event.
        """
        self._internal_state == 'idle'

        if result_callback:
            goal.on('result', result_callback)

        ur_script = goal.goal_message['goal']['script']
        message = Message({'data': ur_script})
        self._urscript_topic.publish(message)

        if timeout:
            self.ros.call_later(
                timeout / 1000., goal._trigger_timeout)


def direct_ur_movel(ros_client, callback, frames, acceleration=None, velocity=None, time=None, radius=None):

    action_client = DirectUrActionClient(ros_client, timeout=50000)

    script_lines = []
    for frame in frames:
        ptp = URPoseTrajectoryPoint(URPose.from_frame(frame), acceleration, velocity, time, radius)
        move = URMovel(ptp)
        script_lines.append(move)

    urgoal = URGoal(script_lines)

    goal = Goal(action_client, Message(urgoal.msg))
    action_client.on('timeout', lambda: print('CLIENT TIMEOUT'))
    # goal.on('feedback', lambda feedback: print(feedback))
    goal.on('result', callback)
    action_client.send_goal(goal)



from __future__ import absolute_import

from .std_msgs import ROSmsg


class ObjectType(ROSmsg):
    """https://docs.ros.org/kinetic/api/object_recognition_msgs/html/msg/ObjectType.html
    """

    def __init__(self, key='key', db='db'):
        self.key = key
        self.db = db

    @classmethod
    def from_msg(cls, msg):
        return cls(msg['key'], msg['db'])



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from compas_fab.backends.interfaces import RemoveCollisionMesh
from compas_fab.backends.pybullet.utils import LOG
from compas_fab.utilities import LazyLoader

pybullet = LazyLoader('pybullet', globals(), 'pybullet')


__all__ = [
    'PyBulletRemoveCollisionMesh',
]


class PyBulletRemoveCollisionMesh(RemoveCollisionMesh):
    """Callable to remove a collision mesh from the planning scene."""
    def __init__(self, client):
        self.client = client

    def remove_collision_mesh(self, id, options=None):
        """Remove a collision mesh from the planning scene.

        Parameters
        ----------
        id : str
            Name of collision mesh to be removed.
        options : dict, optional
            Unused parameter.

        Returns
        -------
        ``None``
        """
        if id not in self.client.collision_objects:
            LOG.warning("Collision object with name '{}' does not exist in scene.".format(id))
            return

        for body_id in self.client.collision_objects[id]:
            pybullet.removeBody(body_id)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging
import os
import sys
from contextlib import contextmanager


__all__ = [
    'LOG',
    'redirect_stdout',
]


@contextmanager
def redirect_stdout(to=os.devnull, enabled=True):
    """Context manager to capture and redirect console output.
    https://stackoverflow.com/a/17954769
    Parameters
    ----------
    to : Location to redirect output to.  Defaults to ``os.devnull``.
    enabled : (:obj:`bool`) Flag to enable or disable redirection.

    Examples
    --------
    >>> import os                                                        # doctest: +SKIP
    >>> with redirect_stdout(to='filename'):                             # doctest: +SKIP
    ...     print("from Python")                                         # doctest: +SKIP
    ...     os.system("echo non-Python applications are also supported") # doctest: +SKIP
    """
    def _redirect_stdout(to_):
        sys.stdout.close()  # + implicit flush()
        os.dup2(to_.fileno(), fd)
        sys.stdout = os.fdopen(fd, 'w')

    # Pytest interferes with file descriptor capture.
    called_from_test = 'pytest' in sys.modules
    enabled = False if called_from_test else enabled

    if not enabled:
        yield
    else:
        fd = sys.stdout.fileno()
        with os.fdopen(os.dup(fd), 'w') as old_stdout:
            with open(to, 'w') as file:
                _redirect_stdout(to_=file)
            try:
                yield
            finally:
                _redirect_stdout(to_=old_stdout)  # restore stdout. buffering and flags such as CLOEXEC may be different


def get_logger(name):
    logger = logging.getLogger(name)

    try:
        from colorlog import ColoredFormatter
        formatter = ColoredFormatter("%(log_color)s%(levelname)-8s%(reset)s %(white)s%(message)s",
                                     datefmt=None,
                                     reset=True,
                                     log_colors={'DEBUG': 'cyan', 'INFO': 'green',
                                                 'WARNING': 'yellow',
                                                 'ERROR': 'red', 'CRITICAL': 'red',
                                                 }
                                     )
    except ImportError:
        formatter = logging.Formatter('[%(levelname)s] %(message)s')

    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)

    return logger


LOG = get_logger(__name__)



import argparse
import glob
import os

from compas._os import create_symlinks
from compas._os import remove_symlinks
from compas_ghpython import get_grasshopper_library_path
from compas_ghpython import get_grasshopper_userobjects_path
from compas_rhino import _check_rhino_version


def get_version_from_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('-v', '--version', choices=['5.0', '6.0', '7.0'], default='6.0')
    args = parser.parse_args()
    return _check_rhino_version(args.version)


def install():
    """Installs the Grasshopper components library."""
    results = []

    try:
        version = get_version_from_args()
        grasshopper_library_path = get_grasshopper_library_path(version)

        for f in _remove_old_ghpy_components(grasshopper_library_path):
            results.append(('compas_fab', 'Uninstalled old GHPY component: {}'.format(f), True))

        dstdir = get_grasshopper_userobjects_path(version)
        srcdir = os.path.dirname(__file__)
        userobjects = glob.glob(os.path.join(srcdir, '*.ghuser'))

        symlinks_to_remove = []
        symlinks_to_add = []
        for src in userobjects:
            dst = os.path.join(dstdir, os.path.basename(src))
            symlinks_to_remove.append(dst)
            symlinks_to_add.append((src, dst))

        remove_symlinks(symlinks_to_remove)
        create_symlinks(symlinks_to_add)

        results.append(('compas_fab', 'Installed {} GH User Objects'.format(len(userobjects)), True))
    except PermissionError:
        raise Exception('Please close first all instances of Rhino and then rerun the command')

    return results


def uninstall():
    """Uninstalls the Grasshopper components library."""
    results = []

    try:
        version = get_version_from_args()
        grasshopper_library_path = get_grasshopper_library_path(version)

        for f in _remove_old_ghpy_components(grasshopper_library_path):
            results.append(('compas_fab', 'Uninstalled old component: {}'.format(f), True))

        dstdir = get_grasshopper_userobjects_path(version)
        srcdir = os.path.dirname(__file__)
        userobjects = glob.glob(os.path.join(srcdir, '*.ghuser'))

        symlinks = []
        for src in userobjects:
            dst = os.path.join(dstdir, os.path.basename(src))
            symlinks.append(dst)

        remove_symlinks(symlinks)

        results.append(('compas_fab', 'Uninstalled {} GH User Objects'.format(len(userobjects)), True))
    except PermissionError:
        raise Exception('Please close first all instances of Rhino and then rerun the command')

    return results


def _remove_old_ghpy_components(grasshopper_library_path):
    oldlibs = glob.glob('%s*.ghpy' % os.path.join(grasshopper_library_path, 'COMPAS FAB_0.1'))
    for f in oldlibs:
        os.remove(f)
        yield f


if __name__ == '__main__':
    install()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import itertools
import os

from compas.geometry import Frame
from compas.robots import Inertia
from compas.robots import Inertial
from compas.robots import Joint
from compas.robots import Mass
from compas.robots import Origin

from compas_fab.backends.interfaces import AddAttachedCollisionMesh
from compas_fab.utilities import LazyLoader

pybullet = LazyLoader('pybullet', globals(), 'pybullet')


__all__ = [
    'PyBulletAddAttachedCollisionMesh',
]


class PyBulletAddAttachedCollisionMesh(AddAttachedCollisionMesh):
    """Callable to add a collision mesh and attach it to the robot."""
    def __init__(self, client):
        self.client = client

    def add_attached_collision_mesh(self, attached_collision_mesh, options=None):
        """Add a collision mesh and attach it to the robot.

        Parameters
        ----------
        attached_collision_mesh : :class:`compas_fab.robots.AttachedCollisionMesh`
            Object containing the collision mesh to be attached.
        options : dict
            Dictionary containing the following key-value pairs:

            - ``"robot"``: (:class:`compas_fab.robots.Robot`) Robot instance
              to which the object should be attached.
            - ``"mass"``: (:obj:`float`) The mass of the attached collision
              object.  Defaults to ``1``.
            - ``inertia"``: (:obj:`list`) The elements of the inertia matrix
              of the attached collision object given as
              ``[<ixx>, <ixy>, <ixz>, <iyy>, <iyz>, <izz>]``.  Defaults to
              ``[1., 0., 0., 1., 0., 1.]``.
            - ``"inertial_origin"``: (:class:`compas.geometry.Frame`) This is
              the pose of the inertial reference frame, relative to the link
              reference frame. Defaults to
              :class:`compas.geometry.Frame.worldXY()`.
            - ``"collision_origin"``(:class:`compas.geometry.Frame`) This is
              the pose of the collision reference frame, relative to the link
              reference frame. Defaults to
              :class:`compas.geometry.Frame.worldXY()`.

        Returns
        -------
        ``None``
        """
        robot = options['robot']
        self.client.ensure_cached_robot_geometry(robot)

        mass = options.get('mass', 1.)
        inertia = options.get('inertia', [1., 0., 0., 1., 0., 1.])
        inertial_origin = options.get('inertial_origin', Frame.worldXY())
        collision_origin = options.get('collision_origin', Frame.worldXY())

        cached_robot_model = self.client.get_cached_robot(robot)

        # add link
        mesh = attached_collision_mesh.collision_mesh.mesh
        name = attached_collision_mesh.collision_mesh.id
        mesh_file_name = name + '.stl'
        mesh_fp = os.path.join(self.client._cache_dir.name, mesh_file_name)
        mesh.to_stl(mesh_fp, binary=True)
        link = cached_robot_model.add_link(name, visual_meshes=[mesh], collision_meshes=[mesh])
        mass_urdf = Mass(mass)
        inertia_urdf = Inertia(*inertia)
        inertial_origin_urdf = Origin(inertial_origin.point, inertial_origin.xaxis, inertial_origin.yaxis)
        inertial_urdf = Inertial(inertial_origin_urdf, mass_urdf, inertia_urdf)
        link.inertial = inertial_urdf
        collision_origin_urdf = Origin(collision_origin.point, collision_origin.xaxis, collision_origin.yaxis)
        for element in itertools.chain(link.visual, link.collision):
            element.geometry.shape.filename = mesh_fp
            element.origin = collision_origin_urdf

        # add joint
        parent_link = cached_robot_model.get_link_by_name(attached_collision_mesh.link_name)
        cached_robot_model.add_joint(name + '_fixed_joint', Joint.FIXED, parent_link, link)

        self.client.reload_from_cache(robot)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from compas.files import XML

__all__ = [
    'RobotSemantics',
]


class RobotSemantics(object):
    """Represents semantic information of a robot.

    The semantic model is based on the
    `Semantic Robot Description Format` (`SRDF`_).

    References
    ----------

    .. _SRDF: https://wiki.ros.org/srdf

    """

    def __init__(self, robot_model,
                 groups=None, main_group_name=None, passive_joints=None,
                 end_effectors=None, disabled_collisions=None,
                 group_states=None):
        self.robot_model = robot_model

        self.groups = groups or {}
        self.main_group_name = main_group_name
        self.passive_joints = passive_joints or []
        self.end_effectors = end_effectors or []
        self.disabled_collisions = disabled_collisions or set()
        self.group_states = group_states or {}

    @property
    def group_names(self):
        return list(self.groups.keys())

    @classmethod
    def from_srdf_file(cls, file, robot_model):
        """Create an instance of semantics based on an SRDF file path or file-like object."""
        xml = XML.from_file(file)
        return cls.from_xml(xml, robot_model)

    @classmethod
    def from_srdf_string(cls, text, robot_model):
        """Create an instance of semantics based on an SRDF string."""
        xml = XML.from_string(text)
        return cls.from_xml(xml, robot_model)

    @classmethod
    def from_xml(cls, xml, robot_model):
        """Create an instance of semantics based on an XML object."""
        groups = _get_groups(xml.root, robot_model)
        passive_joints = _get_passive_joints(xml.root)
        end_effectors = _get_end_effectors(xml.root)
        disabled_collisions = _get_disabled_collisions(xml.root)
        group_states = _get_group_states(xml.root)

        groups_by_links_len = sorted(groups.items(), key=lambda group: len(group[1]['links']))
        main_group_name = groups_by_links_len[-1][0]

        return cls(
            robot_model,
            groups=groups,
            main_group_name=main_group_name,
            passive_joints=passive_joints,
            end_effectors=end_effectors,
            disabled_collisions=disabled_collisions,
            group_states=group_states,
        )

    def get_end_effector_link_name(self, group=None):
        if not group:
            group = self.main_group_name
        return self.groups[group]["links"][-1]

    def get_base_link_name(self, group=None):
        if not group:
            group = self.main_group_name
        return self.groups[group]["links"][0]

    def get_all_configurable_joints(self):
        joints = []
        for joint in self.robot_model.get_configurable_joints():
            if joint.name not in self.passive_joints:
                joints.append(joint)
        return joints

    def get_configurable_joints(self, group=None):
        if not group:
            group = self.main_group_name
        joints = []
        for name in self.groups[group]["joints"]:
            joint = self.robot_model.get_joint_by_name(name)
            if joint:
                if joint.is_configurable() and name not in self.passive_joints:
                    joints.append(joint)
        return joints

    def get_configurable_joint_names(self, group=None):
        return [joint.name for joint in self.get_configurable_joints(group)]


# XML parsing methods
def _get_groups(root, robot_model):
    groups = {}

    for group in root.findall('group'):
        name = group.attrib['name']
        groups[name] = dict(
            links=_get_group_link_names(group, root, robot_model),
            joints=_get_group_joint_names(group, root, robot_model),
        )

    return groups


def _get_group_states(root):
    group_states = {}

    for group_state in root.findall('group_state'):
        group_state_name = group_state.attrib['name']
        group_name = group_state.attrib['group']
        joint_dict = {joint.attrib['name']: joint.attrib['value'] for joint in group_state.findall('joint')}
        group_states.setdefault(group_name, {})[group_state_name] = joint_dict

    return group_states


def _get_group_link_names(group, root, robot_model):
    link_names = []
    for link in group.findall('link'):
        name = link.attrib['name']
        if name not in link_names:
            link_names.append(name)

    for chain in group.findall('chain'):
        for link in robot_model.iter_link_chain(chain.attrib['base_link'], chain.attrib['tip_link']):
            if link.name not in link_names:
                link_names.append(link.name)

    for joint in group.findall('joint'):
        joint = robot_model.get_joint_by_name(joint.attrib['name'])
        if joint:
            name = joint.parent.link
            if name not in link_names:
                link_names.append(name)

    for subgroup in group.findall('group'):
        if subgroup.attrib['name'] != group.attrib['name']:
            # find group element at top level
            for top_group_elem in root.findall('group'):
                if top_group_elem.attrib['name'] == subgroup.attrib['name']:
                    subgroup_link_names = _get_group_link_names(top_group_elem, root, robot_model)
                    for name in subgroup_link_names:
                        if name not in link_names:
                            link_names.append(name)
    return link_names


def _get_group_joint_names(group, root, robot_model):
    joint_names = []
    for link in group.findall('link'):
        link = robot_model.get_link_by_name(link.attrib['name'])
        for joint in link.joints:
            if joint.name not in joint_names:
                joint_names.append(joint.name)

    for chain in group.findall('chain'):
        for joint in robot_model.iter_joint_chain(chain.attrib['base_link'], chain.attrib['tip_link']):
            if joint.name not in joint_names:
                joint_names.append(joint.name)

    for joint in group.findall('joint'):
        if joint.attrib['name'] not in joint_names:
            joint_names.append(joint.attrib['name'])

    for subgroup in group.findall('group'):
        if subgroup.attrib['name'] != group.attrib['name']:
            # find group element at top level
            top_group_elem = _get_group_elem_by_name(subgroup.attrib['name'], root)
            subgroup_joint_names = _get_group_joint_names(top_group_elem, root, robot_model)
            for name in subgroup_joint_names:
                if name not in joint_names:
                    joint_names.append(name)
    return joint_names


def _get_group_elem_by_name(group_name, root):
    for group_elem in root.findall('group'):
        if group_elem.attrib['name'] == group_name:
            return group_elem


def _get_passive_joints(root):
    return [joint.attrib['name'] for joint in root.iter('passive_joint')]


def _get_end_effectors(root):
    return [ee.attrib['parent_link'] for ee in root.findall('end_effector')]


def _get_disabled_collisions(root):
    return {
        tuple([dc.attrib['link1'], dc.attrib['link2']])
        for dc in root.iter('disable_collisions')
    }



class SerialSensor(object):
    """Base class for all sensors with a serial
    interface.

    Args:
        serial (:obj:`serial.Serial`): Instance of the serial
            port used to communicate with the sensor.

    Examples:
        >>> from serial import Serial                           # doctest: +SKIP
        >>> with Serial('COM5', 57600, timeout=1) as serial:    # doctest: +SKIP
        ...     sensor = SerialSensor(serial)                   # doctest: +SKIP
    """
    def __init__(self, serial):
        self.serial = serial

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass



"""
Load robot directly from ROS.

COMPAS FAB v0.18.2
"""
from compas_ghpython.artists import RobotModelArtist
from ghpythonlib.componentbase import executingcomponent as component
from scriptcontext import sticky as st

from compas_fab.ghpython.components import create_id


class ROSRobot(component):
    def RunScript(self, ros_client, load):
        key = create_id(self, 'robot')

        if ros_client and ros_client.is_connected and load:
            # Load URDF from ROS
            st[key] = ros_client.load_robot(load_geometry=True, precision='12f')
            st[key].artist = RobotModelArtist(st[key].model)

        robot = st.get(key, None)
        if robot:  # client sometimes need to be restarted, without needing to reload geometry
            robot.client = ros_client
        return robot



from __future__ import absolute_import

from .std_msgs import ROSmsg
from .std_msgs import Header


class JointState(ROSmsg):
    """https://docs.ros.org/kinetic/api/sensor_msgs/html/msg/JointState.html
    """

    def __init__(self, header=None, name=None, position=None, velocity=None,
                 effort=None):
        self.header = header if header else Header()
        self.name = name if name else []
        self.position = position if position else []
        self.velocity = velocity if velocity else []
        self.effort = effort if effort else []

    @classmethod
    def from_name_and_position(cls, name, position):
        return cls(Header(), name, position, [], [])

    @classmethod
    def from_configuration(cls):
        pass

    @property
    def configuration(self):
        pass

    @classmethod
    def from_msg(cls, msg):
        header = Header.from_msg(msg['header'])
        name = msg['name']
        position = msg['position']
        velocity = msg['velocity']
        effort = msg['effort']
        return cls(header, name, position, velocity, effort)


class MultiDOFJointState(ROSmsg):
    """https://docs.ros.org/kinetic/api/sensor_msgs/html/msg/MultiDOFJointState.html
    """

    def __init__(self, header=None, joint_names=None, transforms=None, twist=None,
                 wrench=None):
        self.header = header if header else Header()
        self.joint_names = joint_names if joint_names else []
        self.transforms = transforms if transforms else []
        self.twist = twist if twist else []
        self.wrench = wrench if wrench else []



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import sys

from compas.robots import Joint
from compas_fab.backends.interfaces import InverseKinematics
from compas_fab.backends.pybullet.conversions import pose_from_frame
from compas_fab.backends.pybullet.exceptions import InverseKinematicsError
from compas_fab.utilities import LazyLoader

pybullet = LazyLoader('pybullet', globals(), 'pybullet')


__all__ = [
    'PyBulletInverseKinematics',
]


class PyBulletInverseKinematics(InverseKinematics):
    """Callable to calculate the robot's inverse kinematics for a given frame."""
    def __init__(self, client):
        self.client = client

    def inverse_kinematics(self, robot, frame_WCF, start_configuration=None, group=None, options=None):
        """Calculate the robot's inverse kinematic for a given frame.

        Parameters
        ----------
        robot : :class:`compas_fab.robots.Robot`
            The robot instance for which inverse kinematics is being calculated.
        frame_WCF: :class:`compas.geometry.Frame`
            The frame to calculate the inverse for.
        start_configuration: :class:`compas_fab.robots.Configuration`, optional
            If passed, the inverse will be calculated such that the calculated
            joint positions differ the least from the start_configuration.
            Defaults to the zero configuration.
        group: str, optional
            The planning group used for determining the end effector and labeling
            the ``start_configuration``. Defaults to the robot's main planning group.
        options: dict, optional
            Dictionary containing the following key-value pairs:

            - ``"link_name"``: (:obj:`str`, optional ) Name of the link for which
              to compute the inverse kinematics.  Defaults to the given group's end
              effector.
            - ``"semi-constrained"``: (:obj:`bool`, optional) When ``True``, only the
              position and not the orientation of ``frame_WCF`` will not be considered
              in the calculation.  Defaults to ``False``.
            - ``"enforce_joint_limits"``: (:obj:`bool`, optional) When ``False``, the
              robot's joint limits will be ignored in the calculation.  Defaults to
              ``True``.

        Returns
        -------
        :obj:`tuple` of :obj:`list`
            A tuple of 2 elements containing a list of joint positions and a list of matching joint names.

        Raises
        ------
        :class:`compas_fab.backends.InverseKinematicsError`
        """
        options = options or {}
        link_name = options.get('link_name') or robot.get_end_effector_link_name(group)
        cached_robot = self.client.get_cached_robot(robot)
        body_id = self.client.get_uid(cached_robot)
        link_id = self.client._get_link_id_by_name(link_name, cached_robot)
        point, orientation = pose_from_frame(frame_WCF)

        joints = cached_robot.get_configurable_joints()
        joints.sort(key=lambda j: j.attr['pybullet']['id'])
        joint_names = [joint.name for joint in joints]

        if start_configuration:
            start_configuration = self.client.set_robot_configuration(robot, start_configuration, group)

        called_from_test = 'pytest' in sys.modules
        if options.get('enforce_joint_limits', True) and not called_from_test:
            lower_limits = [joint.limit.lower if joint.type != Joint.CONTINUOUS else 0 for joint in joints]
            upper_limits = [joint.limit.upper if joint.type != Joint.CONTINUOUS else 2 * math.pi for joint in joints]
            # I don't know what jointRanges needs to be.  Erwin Coumans knows, but he isn't telling.
            # https://stackoverflow.com/questions/49674179/understanding-inverse-kinematics-pybullet
            # https://docs.google.com/document/d/10sXEhzFRSnvFcl3XxNGhnD4N2SedqwdAvK3dsihxVUA/preview?pru=AAABc7276PI*zazLer2rlZ8tAUI8lF98Kw#heading=h.9i02ojf4k3ve
            joint_ranges = [u - l for u, l in zip(upper_limits, lower_limits)]
            rest_configuration = start_configuration or robot.zero_configuration()
            rest_poses = self._get_rest_poses(joint_names, rest_configuration)

            if options.get('semi-constrained'):
                joint_positions = pybullet.calculateInverseKinematics(
                    body_id,
                    link_id,
                    point,
                    lowerLimits=lower_limits,
                    upperLimits=upper_limits,
                    jointRanges=joint_ranges,
                    restPoses=rest_poses,
                )
            else:
                joint_positions = pybullet.calculateInverseKinematics(
                    body_id,
                    link_id,
                    point,
                    orientation,
                    lowerLimits=lower_limits,
                    upperLimits=upper_limits,
                    jointRanges=joint_ranges,
                    restPoses=rest_poses,
                )
        else:
            if options.get('semi-constrained'):
                joint_positions = pybullet.calculateInverseKinematics(
                    body_id,
                    link_id,
                    point,
                )
            else:
                joint_positions = pybullet.calculateInverseKinematics(
                    body_id,
                    link_id,
                    point,
                    orientation,
                )

        if not joint_positions:
            raise InverseKinematicsError()

        return joint_positions, joint_names

    def _get_rest_poses(self, joint_names, configuration):
        name_value_map = {configuration.joint_names[i]: configuration.joint_values[i] for i in range(len(configuration.joint_names))}
        return [name_value_map[name] for name in joint_names]



"""
*******************************************************************************
compas_fab.backends.pybullet
*******************************************************************************

.. module:: compas_fab.backends.pybullet

Package with functionality to interact with `PyBullet <http://pybullet.org/>`_.

.. autosummary::
    :toctree: generated/

    PyBulletClient
    PyBulletError
    CollisionError
    InverseKinematicsError
    PyBulletPlanner

"""

from __future__ import absolute_import

from .client import *                     # noqa: F401,F403
from .const import *                      # noqa: F401,F403
from .conversions import *                # noqa: F401,F403
from .exceptions import *                 # noqa: F401,F403
from .planner import *                    # noqa: F401,F403
from .utils import *                      # noqa: F401,F403

__all__ = [name for name in dir() if not name.startswith('_')]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import random

from compas.geometry import Frame
from compas.geometry import Sphere
from compas.geometry import Transformation
from compas.robots import Configuration
from compas.robots import Joint
from compas.robots import RobotModel

from compas_fab.robots.constraints import Constraint
from compas_fab.robots.constraints import JointConstraint
from compas_fab.robots.constraints import OrientationConstraint
from compas_fab.robots.constraints import PositionConstraint

__all__ = [
    'Robot',
]


class Robot(object):
    """Represents a robot.

    This class binds together several building blocks, such as the robot's
    descriptive model, its semantic information and an instance of a backend
    client into a cohesive programmable interface. This representation builds
    upon the model described in the class :class:`compas.robots.RobotModel` of
    the **COMPAS** framework.

    Attributes
    ----------
    model : :class:`RobotModel`
        The robot model, usually created from an URDF structure.
    artist : :class:`compas.robots.base_artist.BaseRobotModelArtist`
        Instance of the artist used to visualize the robot model. Defaults to ``None``.
    semantics : :class:`compas_fab.robots.RobotSemantics`
        The semantic model of the robot. Defaults to ``None``.
    client : :class:`compas_fab.backends.interfaces.ClientInterface`
        The backend client to use for communication,
        e.g. :class:`compas_fab.backends.RosClient`
    attributes : :obj:`dict`
        Named attributes related to the robot instance.
    attached_tool : :class:`compas_fab.robots.Tool`
        Instance of the tool attached to the robot, if any.
    """

    def __init__(self, model, artist=None, semantics=None, client=None):
        self._scale_factor = 1.
        self.model = model
        self.attached_tool = None
        self.artist = artist
        self.semantics = semantics
        self.client = client
        self.attributes = {}

    @property
    def artist(self):
        """:class:`compas.robots.base_artist.BaseRobotModelArtist`: Artist used to visualize robot model."""
        return self._artist

    @artist.setter
    def artist(self, artist):
        self._artist = artist
        if len(self.model.joints) > 0 and len(self.model.links) > 0:
            self.scale(self._scale_factor)
            if self.attached_tool:
                self.artist.attach_tool_model(self.attached_tool.tool_model)

    @classmethod
    def basic(cls, name, joints=None, links=None, materials=None, **kwargs):
        """Create the most basic instance of a robot, based only on name.

        Parameters
        ----------
        name : :obj:`str`
            Name of the robot
        joints : :obj:`list` of :class:`compas.robots.Joint`, optional
            Robot's joints.
        links : :obj:`list` of :class:`compas.robots.Link`, optional
            Robot's links.
        materials : :obj:`list` of :class:`compas.robots.Material`, optional
            Material description of the robot.
        kwargs : :obj:`dict`
            Keyword arguments passed to the :class:`compas.robots.RobotModel`
            `attr` :obj:`dict`. Accessible from `Robot.model.attr`.


        Returns
        -------
        :class:`Robot`
            Newly created instance of a robot.

        Examples
        --------
        >>> robot = Robot.basic('A robot')
        >>> robot.name
        'A robot'
        """
        model = RobotModel(name, joints=joints or [], links=links or [],
                           materials=materials or [], **kwargs)
        return cls(model, None)

    @property
    def name(self):
        """:obj:`str`: Name of the robot, as defined by its model.

        Examples
        --------
        >>> robot.name
        'ur5'
        """
        return self.model.name

    @property
    def group_names(self):
        """:obj:`list` of :obj:`str`: All planning groups of the robot.

        Examples
        --------
        >>> robot.group_names
        ['manipulator', 'endeffector']

        """
        self.ensure_semantics()
        return self.semantics.group_names

    @property
    def main_group_name(self):
        """:obj:`str`: Robot's main planning group."""
        self.ensure_semantics()
        return self.semantics.main_group_name

    @property
    def root_name(self):
        """:obj:`str`: Robot's root name."""
        return self.model.root.name

    @property
    def group_states(self):
        """:obj:`dict` of :obj:`dict`: All group states of the robot.

        Examples
        --------
        >>> sorted(robot.group_states['manipulator'].keys())
        ['home', 'up']

        """
        self.ensure_semantics()
        return self.semantics.group_states

    def get_end_effector_link_name(self, group=None):
        """Get the name of the robot's end effector link.

        Parameters
        ----------
        group : :obj:`str`, optional
            The name of the group. Defaults to the main planning group.

        Returns
        -------
        :obj:`str`

        Examples
        --------
        >>> robot.get_end_effector_link_name()
        'ee_link'
        """
        if not self.semantics:
            return self.model.get_end_effector_link_name()
        else:
            return self.semantics.get_end_effector_link_name(group)

    def get_end_effector_link(self, group=None):
        """Get the robot's end effector link.

        Parameters
        ----------
        group : :obj:`str`, optional
            The name of the planning group. Defaults to the main planning group.

        Returns
        -------
        :class:`compas.robots.Link`

        Examples
        --------
        >>> link = robot.get_end_effector_link()
        >>> link.name
        'ee_link'
        """
        name = self.get_end_effector_link_name(group)
        return self.model.get_link_by_name(name)

    def get_end_effector_frame(self, group=None, full_configuration=None):
        """Get the frame of the robot's end effector.

        Parameters
        ----------
        group : :obj:`str`, optional
            The name of the planning group. Defaults to the main planning group.
        full_configuration : :class:`Configuration`, optional
            The robot's full configuration, i.e. values for all configurable
            joints of the entire robot. Defaults to the all-zero configuration.

        Returns
        -------
        :class:`compas.geometry.Frame`
        """
        if not full_configuration:
            full_configuration = self.zero_configuration()
        return self.model.forward_kinematics(full_configuration, link_name=self.get_end_effector_link_name(group))

    def get_base_link_name(self, group=None):
        """Get the name of the robot's base link.

        Parameters
        ----------
        group : :obj:`str`, optional
            The name of the planning group. Defaults to the main planning group.

        Returns
        -------
        str

        Examples
        --------
        >>> robot.get_base_link_name()
        'base_link'
        """
        if not self.semantics:
            return self.model.get_base_link_name()
        else:
            return self.semantics.get_base_link_name(group)

    def get_base_link(self, group=None):
        """Get the robot's base link.

        Parameters
        ----------
        group : :obj:`str`, optional
            The name of the planning group. Defaults to the main planning group.

        Returns
        -------
        :class:`compas.robots.Link`

        Examples
        --------
        >>> link = robot.get_base_link()
        >>> link.name
        'base_link'
        """
        name = self.get_base_link_name(group)
        return self.model.get_link_by_name(name)

    def get_base_frame(self, group=None, full_configuration=None):
        """Get the frame of the robot's base link, which is the robot's origin frame.

        Parameters
        ----------
        group : :obj:`str`, optional
            The name of the planning group. Defaults to the main planning group.
        full_configuration : :class:`Configuration`, optional
            The robot's full configuration, i.e. values for all configurable
            joints of the entire robot. Defaults to the all-zero configuration.

        Returns
        -------
        :class:`compas.geometry.Frame`
        """
        if not full_configuration:
            full_configuration = self.zero_configuration()
        return self.model.forward_kinematics(full_configuration, link_name=self.get_base_link_name(group))

    def get_link_names(self, group=None):
        """Get the names of the links in the kinematic chain.

        Parameters
        ----------
        group : :obj:`str`, optional
            The name of the planning group. Defaults to the main planning group.

        Returns
        -------
        :obj:`list` of :obj:`str`

        Examples
        --------
        >>> robot.get_link_names('manipulator')
        ['base_link', 'shoulder_link', 'upper_arm_link', 'forearm_link', 'wrist_1_link', 'wrist_2_link', 'wrist_3_link', 'ee_link']
        """
        base_link_name = self.get_base_link_name(group)
        ee_link_name = self.get_end_effector_link_name(group)
        link_names = []
        for link in self.model.iter_link_chain(base_link_name, ee_link_name):
            link_names.append(link.name)
        return link_names

    def get_link_names_with_collision_geometry(self):
        """Get the names of the links with collision geometry.

        Returns
        -------
        :obj:`list` of :obj:`str`

        Examples
        --------
        >>> robot.get_link_names_with_collision_geometry()
        ['base_link', 'shoulder_link', 'upper_arm_link', 'forearm_link', 'wrist_1_link', 'wrist_2_link', 'wrist_3_link', 'ee_link']
        """
        return [link.name for link in self.model.iter_links() if link.collision]

    def get_configurable_joints(self, group=None):
        """Get the robot's configurable joints.

        Parameters
        ----------
        group : :obj:`str`, optional
            The name of the planning group. Defaults to the main planning group.

        Returns
        -------
        :obj:`list` of :class:`compas.robots.Joint`

        Note
        ----
        If semantics is set and no group is passed, it returns all configurable
        joints of all groups.

        Examples
        --------
        >>> joints = robot.get_configurable_joints('manipulator')
        >>> [j.name for j in joints]
        ['shoulder_pan_joint', 'shoulder_lift_joint', 'elbow_joint', 'wrist_1_joint', 'wrist_2_joint', 'wrist_3_joint']
        """
        if self.semantics:
            if group:
                return self.semantics.get_configurable_joints(group)
            else:
                return self.semantics.get_all_configurable_joints()
        else:
            return self.model.get_configurable_joints()

    def get_joint_types_by_names(self, names):
        """Get a list of joint types given a list of joint names.

        Parameters
        ----------
        names : :obj:`list` of :obj:`str`
            The names of the joints.

        Returns
        -------
        :obj:`list` of :attr:`compas.robots.Joint.SUPPORTED_TYPES`
            List of joint types.
        """
        return self.model.get_joint_types_by_names(names)

    def get_joint_by_name(self, name):
        """RGet the joint in the robot model matching the given name.

        Parameters
        ----------
        name : :obj:`str`
            The name of the joint.

        Returns
        -------
        :class:`compas.robots.Joint`
        """
        return self.model.get_joint_by_name(name)

    def get_configurable_joint_names(self, group=None):
        """Get the configurable joint names.

        Parameters
        ----------
        group : :obj:`str`, optional
            The name of the planning group. Defaults to the main planning group.

        Returns
        -------
        :obj:`list` of :obj:`str`

        Note
        ----
        If semantics is set and no group is passed, it returns all configurable
        joints of all groups.

        Examples
        --------
        >>> robot.get_configurable_joint_names('manipulator')
        ['shoulder_pan_joint', 'shoulder_lift_joint', 'elbow_joint', \
        'wrist_1_joint', 'wrist_2_joint', 'wrist_3_joint']
        """
        configurable_joints = self.get_configurable_joints(group)
        return [j.name for j in configurable_joints]

    def get_configurable_joint_types(self, group=None):
        """Get the configurable joint types.

        Parameters
        ----------
        group : :obj:`str`, optional
            The name of the planning group. Defaults to the main planning group.

        Returns
        -------
        :obj:`list` of :attr:`compas.robots.Joint.SUPPORTED_TYPES`

        Note
        ----
        If :attr:`semantics` is set and no group is passed, it returns all
        configurable joint types of all groups.

        Examples
        --------
        >>> robot.get_configurable_joint_types('manipulator')
        [0, 0, 0, 0, 0, 0]
        """
        configurable_joints = self.get_configurable_joints(group)
        return [j.type for j in configurable_joints]

    # ==========================================================================
    # configurations
    # ==========================================================================

    def zero_configuration(self, group=None):
        """Get the zero joint configuration.

        If zero is out of joint limits ``(upper, lower)`` then
        ``(upper + lower) / 2`` is used as joint value.

        Examples
        --------
        >>> robot.zero_configuration('manipulator')
        Configuration((0.000, 0.000, 0.000, 0.000, 0.000, 0.000), (0, 0, 0, 0, 0, 0), \
            ('shoulder_pan_joint', 'shoulder_lift_joint', 'elbow_joint', 'wrist_1_joint', 'wrist_2_joint', 'wrist_3_joint'))
        """
        values = []
        joint_names = []
        joint_types = []
        for joint in self.get_configurable_joints(group):
            if joint.limit and not (0 <= joint.limit.upper and 0 >= joint.limit.lower):
                values.append((joint.limit.upper + joint.limit.lower)/2.)
            else:
                values.append(0)
            joint_names.append(joint.name)
            joint_types.append(joint.type)
        return Configuration(values, joint_types, joint_names)

    def random_configuration(self, group=None):
        """Get a random configuration.

        Parameters
        ----------
        group : :obj:`str`, optional
            The name of the planning group. Defaults to the main planning group.

        Returns
        -------
        :class:`compas.robots.Configuration`

        Note
        ----
        No collision checking is involved, the configuration may be invalid.
        """
        configurable_joints = self.get_configurable_joints(group)
        values = []
        for joint in configurable_joints:
            if joint.limit:
                values.append(joint.limit.lower + (joint.limit.upper - joint.limit.lower) * random.random())
            else:
                values.append(0)
        joint_names = self.get_configurable_joint_names(group)
        joint_types = self.get_joint_types_by_names(joint_names)
        return Configuration(values, joint_types, joint_names)

    def get_group_configuration(self, group, full_configuration):
        """Get the group's configuration.

        Parameters
        ----------
        group : :obj:`str`
            The name of the planning group.
        full_configuration : :class:`Configuration`
            The configuration for all configurable joints of the robot.

        Returns
        -------
        :class:`Configuration`
            The configuration of the group.
        """
        full_configuration = self._check_full_configuration_and_scale(full_configuration)[0]  # adds joint_names to full_configuration and makes copy
        group_joint_names = self.get_configurable_joint_names(group)
        values = [full_configuration[name] for name in group_joint_names]
        return Configuration(values, self.get_configurable_joint_types(group), group_joint_names)

    def merge_group_with_full_configuration(self, group_configuration, full_configuration, group):
        """Get the robot's full configuration by merging a group's configuration with a full configuration.
        The group configuration takes precedence over the full configuration in
        case a joint value is present in both.

        Parameters
        ----------
        group_configuration : :class:`Configuration`
            The configuration for one of the robot's planning groups.
        full_configuration : :class:`Configuration`
            The configuration for all configurable joints of the robot.
        group : :obj:`str`
            The name of the planning group.

        Returns
        -------
        :class:`Configuration`
            A full configuration with values for all configurable joints.

        Raises
        ------
        :exc:`ValueError`
            If the `full_configuration` does not specify positions for all
            configurable joints, or if the `group_configuration` does not
            specify positions for all configurable joints of the given group.
        """
        if not group_configuration.joint_names:
            group_configuration.joint_names = self.get_configurable_joint_names(group)

        full_configuration = self._check_full_configuration_and_scale(full_configuration)[0]  # adds joint_names to full_configuration and makes copy

        full_configuration = full_configuration.merged(group_configuration)
        return full_configuration

    def get_group_names_from_link_name(self, link_name):
        """Get the names of the groups `link_name` belongs to.

        Parameters
        ----------
        link_name : :obj:`str`
            The name of a link

        Returns
        -------
        :obj:`list` of :obj:`str`
           A list of group names.
        """
        group_names = []
        for group in self.group_names:
            if link_name in self.get_link_names(group):
                group_names.append(group)
        return group_names

    def get_position_by_joint_name(self, configuration, joint_name, group=None):
        """Get the position of named joint in given configuration.

        Parameters
        ----------
        configuration : :class:`Configuration`
            The configuration of the configurable joints.
        joint_name : :obj:`str`
            Name of joint.
        group : :obj:`str`, optional
            The name of the planning group. Defaults to the main planning group.

        Returns
        -------
        :obj:`float`
            Joint position for the given joint.

        Raises
        ------
        :exc:`ValueError`
            If the number of joints in the `configuration` parameter does not
            match the configurable joints of the given `group`.
        """
        names = self.get_configurable_joint_names(group)
        if len(names) != len(configuration.joint_values):
            raise ValueError(
                "Please pass a configuration with %d joint_values or specify group" % len(names))
        return configuration.joint_values[names.index(joint_name)]

    def _check_full_configuration_and_scale(self, full_configuration=None):
        """Either create a full configuration or check if the passed full configuration is valid.

        Parameters
        ----------
        full_configuration : :class:`Configuration`, optional
            The full configuration of the whole robot, including values for all configurable joints.

        Returns
        -------
        (:class:`Configuration`, :class:`Configuration`)
            The full configuration and the scaled full configuration
        """
        if not full_configuration:
            configuration = self.zero_configuration()  # with joint_names
        else:
            joint_names = self.get_configurable_joint_names()  # full configuration
            # full_configuration might have passive joints specified as well, we allow this.
            if len(joint_names) > len(full_configuration.joint_values):
                raise ValueError("Please pass a configuration with {} values, for all configurable joints of the robot.".format(len(joint_names)))
            configuration = full_configuration.copy()
            if not configuration.joint_names:
                configuration.joint_names = joint_names
        return configuration, configuration.scaled(1. / self.scale_factor)

    def get_configuration_from_group_state(self, group, group_state):
        """Get a ``Configuration`` from a group's group state.

        Parameters
        ----------
        group : :obj:`str`
            The name of the planning group.
        group_state : :obj:`str
            The name of the ``group_state``.

        Returns
        -------
        :class:`Configuration`
            The configuration specified by the ``group_state``.
        """
        joint_dict = self.group_states[group][group_state]
        group_joint_names = self.get_configurable_joint_names(group)
        group_joint_types = self.get_configurable_joint_types(group)
        values = [joint_dict[name] for name in group_joint_names]
        return Configuration(values, group_joint_types, group_joint_names)

    # ==========================================================================
    # transformations, coordinate frames
    # ==========================================================================

    def transformation_RCF_WCF(self, group=None):
        """Get the transformation from the robot's coordinate system (RCF) to the world coordinate system (WCF).

        Parameters
        ----------
        group : :obj:`str`, optional
            The name of the planning group. Defaults to the main planning group.

        Returns
        -------
        :class:`compas.geometry.Transformation`

        """
        base_frame = self.get_base_frame(group)
        return Transformation.from_change_of_basis(base_frame, Frame.worldXY())

    def transformation_WCF_RCF(self, group=None):
        """Get the transformation from the world coordinate system (WCF) to the robot's coordinate system (RCF).

        Parameters
        ----------
        group : :obj:`str`, optional
            The name of the planning group. Defaults to the main planning group.

        Returns
        -------
        :class:`compas.geometry.Transformation`

        """
        base_frame = self.get_base_frame(group)
        return Transformation.from_change_of_basis(Frame.worldXY(), base_frame)

    def set_RCF(self, robot_coordinate_frame, group=None):
        """Move the origin frame of the robot to the robot_coordinate_frame.

        Raises
        ------
        :exc:`NotImplementedError`
            Not implemented yet.
        """
        # TODO: must be applied to the model, so that base_frame is RCF
        # Problem: check if conversion wcf/rcf still works with backend
        raise NotImplementedError

    def get_RCF(self, group=None):
        """Get the origin frame of the robot.

        Parameters
        ----------
        group : :obj:`str`, optional
            The name of the planning group. Defaults to the main planning group.

        Returns
        -------
        :class:`compas.geometry.Frame`
            Origin frame of the robot.
        """
        return self.get_base_frame(group)

    def to_local_coordinates(self, frame_WCF, group=None):
        """Represent a frame from the world coordinate system (WCF) in the robot's coordinate system (RCF).

        Parameters
        ----------
        frame_WCF : :class:`compas.geometry.Frame`
            A frame in the world coordinate frame.
        group : :obj:`str`, optional
            The name of the planning group. Defaults to the main planning group.

        Returns
        -------
        :class:`compas.geometry.Frame`
            A frame in the robot's coordinate frame.

        Examples
        --------
        >>> frame_WCF = Frame([-0.363, 0.003, -0.147], [0.388, -0.351, -0.852], [0.276, 0.926, -0.256])
        >>> frame_RCF = robot.to_local_coordinates(frame_WCF)
        >>> frame_RCF
        Frame(Point(-0.363, 0.003, -0.147), Vector(0.388, -0.351, -0.852), Vector(0.276, 0.926, -0.256))
        """
        frame_RCF = frame_WCF.transformed(self.transformation_WCF_RCF(group))
        return frame_RCF

    def to_world_coordinates(self, frame_RCF, group=None):
        """Represent a frame from the robot's coordinate system (RCF) in the world coordinate system (WCF).

        Parameters
        ----------
        frame_RCF : :class:`compas.geometry.Frame`
            A frame in the robot's coordinate frame.
        group : :obj:`str`, optional
            The name of the planning group. Defaults to the main planning group.

        Returns
        -------
        :class:`compas.geometry.Frame`
            A frame in the world coordinate frame.

        Examples
        --------
        >>> frame_RCF = Frame([-0.363, 0.003, -0.147], [0.388, -0.351, -0.852], [0.276, 0.926, -0.256])
        >>> frame_WCF = robot.to_world_coordinates(frame_RCF)
        >>> frame_WCF
        Frame(Point(-0.363, 0.003, -0.147), Vector(0.388, -0.351, -0.852), Vector(0.276, 0.926, -0.256))
        """
        frame_WCF = frame_RCF.transformed(self.transformation_RCF_WCF(group))
        return frame_WCF

    def from_tcf_to_t0cf(self, frames_tcf):
        """Convert a list of frames at the robot's tool tip (tcf frame) to frames at the robot's flange (tool0 frame) using the attached tool.

        Parameters
        ----------
        frames_tcf : :obj:`list` of :class:`compas.geometry.Frame`
            Frames (in WCF) at the robot's tool tip (tcf).

        Returns
        -------
        :obj:`list` of :class:`compas.geometry.Frame`
            Frames (in WCF) at the robot's flange (tool0).

        Raises
        ------
        Exception
            If the attached tool is not set.

        Examples
        --------
        >>> mesh = Mesh.from_stl(compas_fab.get('planning_scene/cone.stl'))
        >>> frame = Frame([0.14, 0, 0], [0, 1, 0], [0, 0, 1])
        >>> robot.attach_tool(Tool(mesh, frame))
        >>> frames_tcf = [Frame((-0.309, -0.046, -0.266), (0.276, 0.926, -0.256), (0.879, -0.136, 0.456))]
        >>> robot.from_tcf_to_t0cf(frames_tcf)
        [Frame(Point(-0.363, 0.003, -0.147), Vector(0.388, -0.351, -0.852), Vector(0.276, 0.926, -0.256))]
        """
        if not self.attached_tool:
            raise Exception("Please attach a tool first.")
        return self.attached_tool.from_tcf_to_t0cf(frames_tcf)

    def from_t0cf_to_tcf(self, frames_t0cf):
        """Convert frames at the robot's flange (tool0 frame) to frames at the robot's tool tip (tcf frame) using the attached tool.

        Parameters
        ----------
        frames_t0cf : :obj:`list` of :class:`compas.geometry.Frame`
            Frames (in WCF) at the robot's flange (tool0).

        Returns
        -------
        :obj:`list` of :class:`compas.geometry.Frame`
            Frames (in WCF) at the robot's tool tip (tcf).

        Raises
        ------
        :exc:`Exception`
            If the robot has no attached tool defined.

        Examples
        --------
        >>> mesh = Mesh.from_stl(compas_fab.get('planning_scene/cone.stl'))
        >>> frame = Frame([0.14, 0, 0], [0, 1, 0], [0, 0, 1])
        >>> robot.attach_tool(Tool(mesh, frame))
        >>> frames_t0cf = [Frame((-0.363, 0.003, -0.147), (0.388, -0.351, -0.852), (0.276, 0.926, -0.256))]
        >>> robot.from_t0cf_to_tcf(frames_t0cf)
        [Frame(Point(-0.309, -0.046, -0.266), Vector(0.276, 0.926, -0.256), Vector(0.879, -0.136, 0.456))]
        """
        if not self.attached_tool:
            raise Exception("Please attach a tool first.")
        return self.attached_tool.from_t0cf_to_tcf(frames_t0cf)

    def attach_tool(self, tool, group=None, touch_links=None):
        """Attach a tool to the robot independently of the model definition.

        Parameters
        ----------
        tool : :class:`Tool`
            The tool that should be attached to the robot's flange.
        group : :obj:`str`, optional
            The planning group to attach this tool to. Defaults to the main
            planning group.
        touch_links : :obj:`list` of :obj:`str`, optional
            A list of link names the end-effector is allowed to touch. Defaults
            to the end-effector link.

        Returns
        -------
        ``None``

        See Also
        --------
        * :meth:`detach_tool`

        Examples
        --------
        >>> mesh = Mesh.from_stl(compas_fab.get('planning_scene/cone.stl'))
        >>> frame = Frame([0.14, 0, 0], [0, 1, 0], [0, 0, 1])
        >>> tool = Tool(mesh, frame)
        >>> robot.attach_tool(tool)
        """
        if not tool.link_name:
            group = group or self.main_group_name
            tool.link_name = self.get_end_effector_link_name(group)
        tool.update_touch_links(touch_links)
        self.attached_tool = tool
        if self.artist:
            self.artist.attach_tool_model(tool.tool_model)

    def detach_tool(self):
        """Detach the attached tool.

        See Also
        --------
        * :meth:`attach_tool`
        """
        self.attached_tool = None
        if self.artist:
            self.artist.detach_tool_model()

    # ==========================================================================
    # checks
    # ==========================================================================

    def ensure_client(self):
        """Check if the client is set.
       Raises
       ------
       :exc:`Exception`
           If :attr:`client` is not set
       """
        if not self.client:
            raise Exception(
                'This method is only callable once a client is assigned.')

    def ensure_semantics(self):
        """Check if semantics is set.
        Raises
        ------
        :exc:`Exception`
            If :attr:`semantics` is not set.
        """
        if not self.semantics:
            raise Exception(
                'This method is only callable once a semantic model is assigned.')

    def ensure_geometry(self):
        """Check if the model's geometry has been loaded.
        Raises
        ------
        :exc:`Exception`
            If geometry has not been loaded.
        """
        self.model.ensure_geometry()
    # ==========================================================================
    # constraints
    # ==========================================================================

    def orientation_constraint_from_frame(self, frame_WCF, tolerances_axes,
                                          group=None):
        r"""Create an orientation constraint from a frame on the group's end-effector link.

        Parameters
        ----------
        frame_WCF: :class:`compas.geometry.Frame`
            The frame from which we create the orientation constraint.
        tolerances_axes: :obj:`list` of :obj:`float`
            Error tolerances t\ :sub:`i` for each of the frame's axes in
            radians. If only one value is passed it will be uses for all 3 axes.
        group: :obj:`str`, optional
            The planning group for which we specify the constraint. Defaults to
            the robot's main planning group.

        Returns
        -------
        :class:`OrientationConstraint`

        Raises
        ------
        :exc:`ValueError`
            If tolerance axes given are not one or three values.

        Notes
        -----
        The rotation tolerance for an axis is defined by the other vector
        component values for rotation around corresponding axis.
        If you specify the tolerances_axes vector with ``[0.01, 0.01, 6.3]``, it
        means that the frame's x-axis and y-axis are allowed to rotate about the
        z-axis by an angle of 6.3 radians, whereas the z-axis would only rotate
        by 0.01.


        Examples
        --------
        >>> frame = Frame([0.4, 0.3, 0.4], [0, 1, 0], [0, 0, 1])
        >>> tolerances_axes = [math.radians(1)] * 3
        >>> group = robot.main_group_name
        >>> robot.orientation_constraint_from_frame(frame, tolerances_axes, group=group)
        OrientationConstraint('ee_link', [0.5, 0.5, 0.5, 0.5], [0.017453292519943295, 0.017453292519943295, 0.017453292519943295], 1.0)
        """

        ee_link = self.get_end_effector_link_name(group)

        tolerances_axes = list(tolerances_axes)
        if len(tolerances_axes) == 1:
            tolerances_axes *= 3
        elif len(tolerances_axes) != 3:
            raise ValueError("Must give either one or 3 values")
        return OrientationConstraint(ee_link, frame_WCF.quaternion, tolerances_axes)

    def position_constraint_from_frame(self, frame_WCF, tolerance_position, group=None):
        """Create a position constraint from a frame on the group's end-effector link.

        Parameters
        ----------
        frame_WCF : :class:`compas.geometry.Frame`
            The frame from which we create position and orientation constraints.
        tolerance_position : :obj:`float`
            The allowed tolerance to the frame's position (defined in the
            robot's units).
        group: :obj:`str`, optional
            The planning group for which we specify the constraint. Defaults to
            the robot's main planning group.

        Returns
        -------
        :class:`PositionConstraint`

        See Also
        --------
        * :meth:`PositionConstraint.from_box`
        * :meth:`PositionConstraint.from_mesh`
        * :meth:`PositionConstraint.from_sphere`

        Notes
        -----
        The rotation tolerance for an axis is defined by the other vector
        component values for rotation around corresponding axis.
        If you specify the tolerances_axes vector with ``[0.01, 0.01, 6.3]``, it
        means that the frame's x-axis and y-axis are allowed to rotate about the
        z-axis by an angle of 6.3 radians, whereas the z-axis would only rotate
        by 0.01.

        Examples
        --------
        >>> frame = Frame([0.4, 0.3, 0.4], [0, 1, 0], [0, 0, 1])
        >>> tolerance_position = 0.001
        >>> robot.position_constraint_from_frame(frame, tolerance_position)
        PositionConstraint('ee_link', BoundingVolume(2, Sphere(Point(0.400, 0.300, 0.400), 0.001)), 1.0)
        """

        ee_link = self.get_end_effector_link_name(group)
        sphere = Sphere(frame_WCF.point, tolerance_position)
        return PositionConstraint.from_sphere(ee_link, sphere)

    def constraints_from_frame(self, frame_WCF, tolerance_position, tolerances_axes, group=None):
        r"""Create a position and an orientation constraint from a frame calculated for the group's end-effector link.

        Parameters
        ----------
        frame_WCF: :class:`compas.geometry.Frame`
            The frame from which we create position and orientation constraints.
        tolerance_position: :obj:`float`
            The allowed tolerance to the frame's position (defined in the
            robot's units).
        tolerances_axes: :obj:`list` of :obj:`float`
            Error tolerances t\ :sub:`i` for each of the frame's axes in
            radians. If only one value is passed it will be uses for all 3 axes.
        group: :obj:`str`, optional
            The planning group for which we specify the constraint. Defaults to
            the robot's main planning group.

        Returns
        -------
        :obj:`list` of :class:`Constraint`

        See Also
        --------
        * :meth:`PositionConstraint.from_box`
        * :meth:`PositionConstraint.from_mesh`
        * :meth:`PositionConstraint.from_sphere`
        * :meth:`orientation_constraint_from_frame`

        Notes
        -----
        The rotation tolerance for an axis is defined by the other vector
        component values for rotation around corresponding axis.
        If you specify the tolerances_axes vector with ``[0.01, 0.01, 6.3]``, it
        means that the frame's x-axis and y-axis are allowed to rotate about the
        z-axis by an angle of 6.3 radians, whereas the z-axis would only rotate
        by 0.01.


        Examples
        --------
        >>> frame = Frame([0.4, 0.3, 0.4], [0, 1, 0], [0, 0, 1])
        >>> tolerance_position = 0.001
        >>> tolerances_axes = [math.radians(1)]
        >>> group = robot.main_group_name
        >>> robot.constraints_from_frame(frame, tolerance_position, tolerances_axes, group) # doctest: +NORMALIZE_WHITESPACE
        [PositionConstraint('ee_link', BoundingVolume(2, Sphere(Point(0.400, 0.300, 0.400), 0.001)), 1.0),
        OrientationConstraint('ee_link', [0.5, 0.5, 0.5, 0.5], [0.017453292519943295, 0.017453292519943295, 0.017453292519943295], 1.0)]
        """
        pc = self.position_constraint_from_frame(frame_WCF, tolerance_position, group)
        oc = self.orientation_constraint_from_frame(frame_WCF, tolerances_axes, group)
        return [pc, oc]

    def constraints_from_configuration(self, configuration, tolerances_above, tolerances_below, group=None):
        """Create joint constraints for all joints of the configuration.

        Parameters
        ----------
        configuration: :class:`Configuration`
            The target configuration.
        tolerances_above: :obj:`list` of :obj:`float`
            The tolerances above the targeted configuration's joint value on
            each of the joints, defining the upper bound in radians to be
            achieved. If only one value is passed, it will be used to create
            upper bounds for all joint constraints.
        tolerances_below: :obj:`list` of :obj:`float`
            The tolerances below the targeted configuration's joint value on
            each of the joints, defining the upper bound in radians to be
            achieved. If only one value is passed, it will be used to create
            lower bounds for all joint constraints.
        group: :obj:`str`, optional
            The planning group for which we specify the constraint. Defaults to
            the robot's main planning group.

        Returns
        -------
        :obj:`list` of :class:`JointConstraint`

        Raises
        ------
        :exc:`ValueError`
            If the passed configuration does not correspond to the group.
        :exc:`ValueError`
            If the passed list of tolerance values have a different length than
            the configuration.

        Notes
        -----
        Make sure that you are using the correct tolerance units if your robot
        has different joint types defined.

        Examples
        --------
        >>> configuration = Configuration.from_revolute_values([-0.042, 4.295, -4.110, -3.327, 4.755, 0.])
        >>> tolerances_above = [math.radians(1)] * 6
        >>> tolerances_below = [math.radians(1)] * 6
        >>> group = robot.main_group_name
        >>> robot.constraints_from_configuration(configuration, tolerances_above, tolerances_below, group)
        [JointConstraint('shoulder_pan_joint', -0.042, 0.017453292519943295, 0.017453292519943295, 1.0), \
        JointConstraint('shoulder_lift_joint', 4.295, 0.017453292519943295, 0.017453292519943295, 1.0), \
        JointConstraint('elbow_joint', -4.11, 0.017453292519943295, 0.017453292519943295, 1.0), \
        JointConstraint('wrist_1_joint', -3.327, 0.017453292519943295, 0.017453292519943295, 1.0), \
        JointConstraint('wrist_2_joint', 4.755, 0.017453292519943295, 0.017453292519943295, 1.0), \
        JointConstraint('wrist_3_joint', 0.0, 0.017453292519943295, 0.017453292519943295, 1.0)]
        """
        if not group:
            group = self.main_group_name

        joint_names = self.get_configurable_joint_names(group)
        if len(joint_names) != len(configuration.joint_values):
            raise ValueError("The passed configuration has %d joint_values, the group %s needs however: %d" % (
                len(configuration.joint_values), group, len(joint_names)))
        if len(tolerances_above) == 1:
            tolerances_above = tolerances_above * len(joint_names)
        elif len(tolerances_above) != len(configuration.joint_values):
            raise ValueError("The passed configuration has %d joint_values, the tolerances_above however: %d" % (
                len(configuration.joint_values), len(tolerances_above)))
        if len(tolerances_below) == 1:
            tolerances_below = tolerances_below * len(joint_names)
        elif len(tolerances_below) != len(configuration.joint_values):
            raise ValueError("The passed configuration has %d joint_values, the tolerances_below however: %d" % (
                len(configuration.joint_values), len(tolerances_below)))

        constraints = []
        for name, value, tolerance_above, tolerance_below in zip(joint_names, configuration.joint_values, tolerances_above, tolerances_below):
            constraints.append(JointConstraint(name, value, tolerance_above, tolerance_below))
        return constraints

    # ==========================================================================
    # services
    # ==========================================================================

    def inverse_kinematics(self, frame_WCF, start_configuration=None, group=None, return_full_configuration=False, options=None):
        """Calculate the robot's inverse kinematic for a given frame.

        Parameters
        ----------
        frame_WCF : :class:`compas.geometry.Frame`
            The frame to calculate the inverse kinematic for.
        start_configuration : :class:`Configuration`, optional
            If passed, the inverse will be calculated such that the calculated
            joint positions differ the least from the start_configuration.
            Defaults to the zero configuration.
        group: :obj:`str`, optional
            The planning group used for calculation. Defaults to the robot's
            main planning group.
        return_full_configuration : :obj:`bool`, optional
            If ``True``, returns a full configuration with all joint values
            specified, including passive ones if available. Defaults to ``False``.
        options: :obj:`dict`, optional
            Dictionary containing the key-value pairs of additional options.
            The valid options are specific to the backend in use.
            Check the API reference of the IK backend implementation for more details.

        Raises
        ------
        :exc:`compas_fab.backends.BackendError`
            If no configuration can be found.

        Returns
        -------
        :class:`Configuration`
            The planning group's configuration.

        Examples
        --------
        >>> frame_WCF = Frame([0.3, 0.1, 0.5], [1, 0, 0], [0, 1, 0])
        >>> start_configuration = robot.zero_configuration()
        >>> group = robot.main_group_name
        >>> robot.inverse_kinematics(frame_WCF, start_configuration, group)                 # doctest: +SKIP
        Configuration((4.045, 5.130, -2.174, -6.098, -5.616, 6.283), (0, 0, 0, 0, 0, 0))    # doctest: +SKIP
        """
        options = options or {}
        attached_collision_meshes = options.get('attached_collision_meshes') or []

        self.ensure_client()
        group = group or self.main_group_name if self.semantics else None

        start_configuration, start_configuration_scaled = self._check_full_configuration_and_scale(start_configuration)

        frame_WCF_scaled = frame_WCF.copy()
        frame_WCF_scaled.point /= self.scale_factor  # must be in meters

        if self.attached_tool:
            attached_collision_meshes.extend(self.attached_tool.attached_collision_meshes)

        options['attached_collision_meshes'] = attached_collision_meshes

        # The returned joint names might be more than the requested ones if there are passive joints present
        joint_positions, joint_names = self.client.inverse_kinematics(self,
                                                                      frame_WCF_scaled,
                                                                      start_configuration_scaled,
                                                                      group,
                                                                      options)
        if return_full_configuration:
            # build configuration including passive joints, but no sorting
            joint_types = self.get_joint_types_by_names(joint_names)
            configuration = Configuration(joint_positions, joint_types, joint_names)
        else:
            # sort values for group configuration
            joint_state = dict(zip(joint_names, joint_positions))
            group_joint_names = self.get_configurable_joint_names(group)
            values = [joint_state[name] for name in group_joint_names]
            configuration = Configuration(values, self.get_configurable_joint_types(group), group_joint_names)

        return configuration.scaled(self.scale_factor)

    def inverse_kinematics_deprecated(self, frame_WCF, start_configuration=None,
                                      group=None, avoid_collisions=True,
                                      constraints=None, attempts=8,
                                      attached_collision_meshes=None,
                                      return_full_configuration=False):
        return self.inverse_kinematics(frame_WCF,
                                       start_configuration,
                                       group,
                                       return_full_configuration,
                                       options=dict(
                                           avoid_collisions=avoid_collisions,
                                           constraints=constraints,
                                           attempts=attempts,
                                           attached_collision_meshes=attached_collision_meshes,
                                       ))

    def forward_kinematics(self, configuration, group=None, options=None):
        """Calculate the robot's forward kinematic.

        Parameters
        ----------
        configuration: :class:`compas_fab.robots.Configuration`
            The full configuration to calculate the forward kinematic for. If no
            full configuration is passed, the zero-joint state for the other
            configurable joints is assumed.
        group: obj:`str`, optional
            The planning group used for the calculation. Defaults to the robot's
            main planning group.
        options: obj:`dict`, optional
            Dictionary containing the following key-value pairs:

            - ``"solver"``: (:obj:`str`, optional) If ``None`` calculates FK
              with the client if it exists or with the robot model.
              If ``'model'`` use the robot model to calculate FK.
              Other values depend on specific backend implementation, some backends might
              allow selecting different FK solvers dynamically.
            - ``"link"``: (:obj:`str`, optional) The name of the link to
              calculate the forward kinematics for. Defaults to the group's end
              effector link.
              Backwards compatibility note: if there's no ``link`` option, the
              planner will try also ``ee_link`` as fallback before defaulting
              to the end effector's link.

            There are additional options that are specific to the backend in use.
            Check the API reference of the FK backend implementation for more details.

        Returns
        -------
        :class:`compas.geometry.Frame`
            The frame in the world's coordinate system (WCF).

        Raises
        ------
        :exc:`ValueError`
            If `link_name` doesn't match any of the :class:`Robot` instance's links.
        :exc:`NotImplementedError`
            If forward kinematic method for given `backend` is not implemented.

        Examples
        --------
        >>> configuration = Configuration.from_revolute_values([-2.238, -1.153, -2.174, 0.185, 0.667, 0.000])
        >>> group = robot.main_group_name
        >>> frame_WCF_c = robot.forward_kinematics(configuration, group)
        >>> options = {'solver': 'model'}
        >>> frame_WCF_m = robot.forward_kinematics(configuration, group, options)
        >>> frame_WCF_c == frame_WCF_m
        True
        """
        options = options or {}
        solver = options.get('solver')

        group = group or self.main_group_name if self.semantics else None

        full_configuration = self.merge_group_with_full_configuration(configuration, self.zero_configuration(), group)
        full_configuration, full_configuration_scaled = self._check_full_configuration_and_scale(full_configuration)

        # If there's no client, we default to `model` solver if there is no other assigned.
        if not self.client:
            solver = solver or 'model'

        # Solve with the model
        if solver == 'model':
            link = options.get('link', options.get('ee_link'))
            link = link or self.get_end_effector_link_name(group)
            if link not in self.get_link_names(group):
                raise ValueError('Link name {} does not exist in planning group'.format(link))

            frame_WCF = self.model.forward_kinematics(full_configuration, link)

        # Otherwise, pass everything down to the client
        else:
            self.ensure_client()
            frame_WCF = self.client.forward_kinematics(self,
                                                       full_configuration_scaled,
                                                       group,
                                                       options)

        # Scale and return
        frame_WCF.point *= self.scale_factor
        return frame_WCF

    def forward_kinematics_deprecated(self, configuration, group=None, backend=None, ee_link=None):
        return self.forward_kinematics(configuration, group, options=dict(solver=backend, link=ee_link))

    def plan_cartesian_motion(self, frames_WCF, start_configuration=None, group=None, options=None):
        """Calculate a cartesian motion path (linear in tool space).

        Parameters
        ----------
        frames_WCF : :obj:`list` of :class:`compas.geometry.Frame`
            The frames through which the path is defined.
        start_configuration : :class:`Configuration`, optional
            The robot's full configuration, i.e. values for all configurable
            joints of the entire robot, at the starting position. Defaults to
            the all-zero configuration.
        group : :obj:`str`, optional
            The planning group used for calculation. Defaults to the robot's
            main planning group.
        options : :obj:`dict`, optional
            Dictionary containing the following key-value pairs:

            - max_step :: :obj:`float`, optional
                The approximate distance between the
                calculated points. (Defined in the robot's units.) Defaults to ``0.01``.
            - path_constraints :: :obj:`list` of :class:`compas_fab.robots.Constraint`, optional
                Optional constraints that can be imposed along the solution path.
                Note that path calculation won't work if the start_configuration
                violates these constraints. Defaults to ``None``.
            - attached_collision_meshes :: :obj:`list` of :class:`compas_fab.robots.AttachedCollisionMesh`
                Defaults to ``None``.

            There are additional options that are specific to the backend in use.
            Check the API reference of the cartesian motion planner backend implementation
            for more details.

        Returns
        -------
        :class:`JointTrajectory`
            The calculated trajectory.

        Examples
        --------

        >>> robot.client = RosClient()
        >>> robot.client.run()
        >>> frames = [Frame([0.3, 0.1, 0.5], [1, 0, 0], [0, 1, 0]),\
                      Frame([0.5, 0.1, 0.6], [1, 0, 0], [0, 1, 0])]
        >>> start_configuration = Configuration.from_revolute_values([-0.042, 0.033, -2.174, 5.282, -1.528, 0.000])
        >>> group = robot.main_group_name
        >>> options = {'max_step': 0.01,\
                       'jump_threshold': 1.57,\
                       'avoid_collisions': True}
        >>> trajectory = robot.plan_cartesian_motion(frames,\
                                                     start_configuration,\
                                                     group=group,\
                                                     options=options)
        >>> len(trajectory.points) > 1
        True
        >>> robot.client.close()
        """
        options = options or {}
        max_step = options.get('max_step')
        path_constraints = options.get('path_constraints')
        attached_collision_meshes = options.get('attached_collision_meshes') or []

        self.ensure_client()
        if not group:
            group = self.main_group_name  # ensure semantics

        # NOTE: start_configuration has to be a full robot configuration, such
        # that all configurable joints of the whole robot are defined for planning.
        start_configuration, start_configuration_scaled = self._check_full_configuration_and_scale(start_configuration)

        frames_WCF_scaled = []
        for frame in frames_WCF:
            frames_WCF_scaled.append(Frame(frame.point * 1. / self.scale_factor, frame.xaxis, frame.yaxis))

        if path_constraints:
            path_constraints_WCF_scaled = []
            for c in path_constraints:
                cp = c.copy()
                if c.type == Constraint.JOINT:
                    joint = self.get_joint_by_name(c.joint_name)
                    if joint.is_scalable():
                        cp.scale(1. / self.scale_factor)
                else:
                    cp.scale(1. / self.scale_factor)
                path_constraints_WCF_scaled.append(cp)
        else:
            path_constraints_WCF_scaled = None

        if self.attached_tool:
            attached_collision_meshes.extend(self.attached_tool.attached_collision_meshes)

        options['attached_collision_meshes'] = attached_collision_meshes
        options['path_constraints'] = path_constraints
        if max_step:
            options['max_step'] = max_step / self.scale_factor

        trajectory = self.client.plan_cartesian_motion(
            robot=self,
            frames_WCF=frames_WCF_scaled,
            start_configuration=start_configuration_scaled,
            group=group,
            options=options)

        # Scale everything back to robot's scale
        for pt in trajectory.points:
            pt.scale(self.scale_factor)

        trajectory.start_configuration.scale(self.scale_factor)

        return trajectory

    def plan_cartesian_motion_deprecated(self, frames_WCF, start_configuration=None,
                                         max_step=0.01, jump_threshold=1.57,
                                         avoid_collisions=True, group=None,
                                         path_constraints=None,
                                         attached_collision_meshes=None):
        return self.plan_cartesian_motion(frames_WCF,
                                          start_configuration=start_configuration,
                                          group=group,
                                          options=dict(
                                              max_step=max_step,
                                              jump_threshold=jump_threshold,
                                              avoid_collisions=avoid_collisions,
                                              path_constraints=path_constraints,
                                              attached_collision_meshes=attached_collision_meshes,
                                          ))

    def plan_motion(self, goal_constraints, start_configuration=None, group=None, options=None):
        """Calculate a motion path.

        Parameters
        ----------
        goal_constraints : list of :class:`Constraint`
            The goal to be achieved, defined in a set of constraints.
            Constraints can be very specific, for example defining value domains
            for each joint, such that the goal configuration is included,
            or defining a volume in space, to which a specific robot link (e.g.
            the end-effector) is required to move to.
        start_configuration : :class:`.Configuration`, optional
            The robot's full configuration, i.e. values for all configurable
            joints of the entire robot, at the starting position. Defaults to
            the all-zero configuration.
        group : :obj:`str`, optional
            The name of the group to plan for. Defaults to the robot's main
            planning group.
        options : :obj:`dict`, optional
            Dictionary containing the following key-value pairs:

            - path_constraints :: :obj:`list` of :class:`Constraint`, optional
                Optional constraints that can be imposed along the solution path.
                Note that path calculation won't work if the start_configuration
                violates these constraints. Defaults to ``None``.
            - attached_collision_meshes :: :obj:`list` of :class:`compas_fab.robots.AttachedCollisionMesh`
                Defaults to ``None``.

            There are additional options that are specific to the backend in use.
            Check the API reference of the motion planner backend implementation
            for more details.

        Returns
        -------
        :class:`JointTrajectory`
            The calculated trajectory.

        Examples
        --------

        Using position and orientation constraints:

        >>> robot.client = RosClient()
        >>> robot.client.run()
        >>> frame = Frame([0.4, 0.3, 0.4], [0, 1, 0], [0, 0, 1])
        >>> tolerance_position = 0.001
        >>> tolerances_axes = [math.radians(1)] * 3
        >>> start_configuration = Configuration.from_revolute_values([-0.042, 4.295, 0, -3.327, 4.755, 0.])
        >>> group = robot.main_group_name
        >>> goal_constraints = robot.constraints_from_frame(frame, tolerance_position, tolerances_axes, group)
        >>> robot.attached_tool = None
        >>> trajectory = robot.plan_motion(goal_constraints, start_configuration, group, {'planner_id': 'RRTConnectkConfigDefault'})
        >>> trajectory.fraction
        1.0
        >>> len(trajectory.points) > 1
        True
        >>> robot.client.close()

        Using joint constraints (to the UP configuration):

        >>> robot.client = RosClient()
        >>> robot.client.run()
        >>> configuration = Configuration.from_revolute_values([0.0, -1.5707, 0.0, -1.5707, 0.0, 0.0])
        >>> tolerances_above = [math.radians(5)] * len(configuration.joint_values)
        >>> tolerances_below = [math.radians(5)] * len(configuration.joint_values)
        >>> group = robot.main_group_name
        >>> goal_constraints = robot.constraints_from_configuration(configuration, tolerances_above, tolerances_below, group)
        >>> trajectory = robot.plan_motion(goal_constraints, start_configuration, group, {'planner_id': 'RRTConnectkConfigDefault'})
        >>> trajectory.fraction
        1.0
        >>> len(trajectory.points) > 1
        True
        >>> robot.client.close()
        """
        options = options or {}
        path_constraints = options.get('path_constraints')
        attached_collision_meshes = options.get('attached_collision_meshes') or []

        # TODO: for the motion plan request a list of possible goal constraints
        # can be passed, from which the planner will try to find a path that
        # satisfies at least one of the specified goal constraints. For now only
        # one set of goal constraints is supported.

        # TODO: add workspace_parameters

        self.ensure_client()
        if not group:
            group = self.main_group_name

        # NOTE: start_configuration has to be a full robot configuration, such
        # that all configurable joints of the whole robot are defined for planning.
        start_configuration, start_configuration_scaled = self._check_full_configuration_and_scale(start_configuration)

        goal_constraints_WCF_scaled = []
        for c in goal_constraints:
            cp = c.copy()
            if c.type == Constraint.JOINT:
                joint = self.get_joint_by_name(c.joint_name)
                if joint.is_scalable():
                    cp.scale(1. / self.scale_factor)
            else:
                cp.scale(1. / self.scale_factor)
            goal_constraints_WCF_scaled.append(cp)

        # Transform path constraints to RCF and scale
        if path_constraints:
            path_constraints_WCF_scaled = []
            for c in path_constraints:
                cp = c.copy()
                if c.type == Constraint.JOINT:
                    joint = self.get_joint_by_name(c.joint_name)
                    if joint.is_scalable():
                        cp.scale(1. / self.scale_factor)
                else:
                    cp.scale(1. / self.scale_factor)
                path_constraints_WCF_scaled.append(cp)
        else:
            path_constraints_WCF_scaled = None

        if self.attached_tool:
            attached_collision_meshes.extend(self.attached_tool.attached_collision_meshes)

        options['attached_collision_meshes'] = attached_collision_meshes
        options['path_constraints'] = path_constraints_WCF_scaled

        trajectory = self.client.plan_motion(
            robot=self,
            goal_constraints=goal_constraints_WCF_scaled,
            start_configuration=start_configuration_scaled,
            group=group,
            options=options,
        )

        # Scale everything back to robot's scale
        for pt in trajectory.points:
            pt.scale(self.scale_factor)

        trajectory.start_configuration.scale(self.scale_factor)

        return trajectory

    def plan_motion_deprecated(self, goal_constraints, start_configuration=None,
                               group=None, path_constraints=None, planner_id='RRTConnectkConfigDefault',
                               num_planning_attempts=1, allowed_planning_time=2.,
                               max_velocity_scaling_factor=1.,
                               max_acceleration_scaling_factor=1.,
                               attached_collision_meshes=None):
        return self.plan_motion(goal_constraints,
                                start_configuration,
                                group,
                                options=dict(
                                    path_constraints=path_constraints,
                                    planner_id=planner_id,
                                    num_planning_attempts=num_planning_attempts,
                                    allowed_planning_time=allowed_planning_time,
                                    max_velocity_scaling_factor=max_velocity_scaling_factor,
                                    max_acceleration_scaling_factor=max_acceleration_scaling_factor,
                                    attached_collision_meshes=attached_collision_meshes,
                                ))

    def transformed_frames(self, configuration, group=None):
        """Get the robot's transformed frames.

        Parameters
        ----------
        configuration : :class:`Configuration`
            Configuration to compute transformed frames for.
        group : :obj:`str`, optional
            The planning group used for the calculation. Defaults to the robot's
            main planning group.

        Returns
        -------
        :obj:`list` of :class:`compas.geometry.Frame`
            Transformed frames.
        """
        if not len(configuration.joint_names):
            configuration.joint_names = self.get_configurable_joint_names(group)
        return self.model.transformed_frames(configuration)

    def transformed_axes(self, configuration, group=None):
        """Get the robot's transformed axes.

        Parameters
        ----------
        configuration : :class:`Configuration`
            Configuration to compute transformed axes for.
        group : :obj:`str`, optional
            The planning group used for the calculation. Defaults to the robot's
            main planning group.

        Returns
        -------
        :obj:`list` of :class:`compas.geometry.Vector`
            Transformed axes.
        """
        if not len(configuration.joint_names):
            configuration.joint_names = self.get_configurable_joint_names(group)
        return self.model.transformed_axes(configuration)

    # ==========================================================================
    # drawing
    # ==========================================================================

    def update(self, configuration, group=None, visual=True, collision=True):
        """Update the robot's geometry.

        Parameters
        ----------
        configuration : :class:`Configuration`
            Instance of the configuration (joint state) to move to.
        group : :obj:`str`, optional
            The name of the group to plan for. Defaults to the robot's main
            planning group.
        visual : :obj:`bool`, optional
            ``True`` if the visual geometry should be also updated, otherwise ``False``.
            Defaults to ``True``.
        collision : :obj:`bool`, optional
            ``True`` if the collision geometry should be also updated, otherwise ``False``.
            Defaults to ``True``.
        """
        group = group or self.main_group_name if self.semantics else None

        if not len(configuration.joint_names):
            configuration.joint_names = self.get_configurable_joint_names(group)

        self.artist.update(configuration, visual, collision)

    def draw_visual(self):
        """Draw the robot's visual geometry using the defined :attr:`Robot.artist`."""
        return self.artist.draw_visual()

    def draw_collision(self):
        """Draw the robot's collision geometry using the defined :attr:`Robot.artist`."""
        return self.artist.draw_collision()

    def draw(self):
        """Alias of :meth:`draw_visual`."""
        return self.draw_visual()

    # TODO: add artist.draw_attached_tool
    # def draw_attached_tool(self):
    #     """Draw the attached tool using the defined :attr:`Robot.artist`."""
    #     if self.artist and self.attached_tool:
    #         return self.artist.draw_attached_tool()

    def scale(self, factor):
        """Scale the robot geometry by a factor (absolute).

        Parameters
        ----------
        factor : :obj:`float`
            The factor to scale the robot with.

        Returns
        -------
        ``None``
        """
        self.model.scale(factor)
        if self.artist:
            self.artist.scale(factor)
        else:
            self._scale_factor = factor

    @property
    def scale_factor(self):
        """:obj:`float`: Robot's scale factor."""
        if self.artist:
            return self.artist.scale_factor
        else:
            return self._scale_factor

    def info(self):
        """Print information about the robot."""
        print("The robot's name is '%s'." % self.name)
        if self.semantics:
            print("The planning groups are:", self.group_names)
            print("The main planning group is '%s'." % self.main_group_name)
            configurable_joints = self.get_configurable_joints(
                self.main_group_name)
        else:
            configurable_joints = self.get_configurable_joints()
        print("The end-effector's name is '%s'." %
              self.get_end_effector_link_name())
        if self.attached_tool:
            print("The robot has a tool at the %s link attached." % self.attached_tool.link_name)
        else:
            print("The robot has NO tool attached.")
        print("The base link's name is '%s'" % self.get_base_link_name())
        print("The base_frame is:", self.get_base_frame())
        print("The robot's joints are:")
        for joint in configurable_joints:
            info = "\t* '%s' is of type '%s'" % (
                joint.name, list(Joint.SUPPORTED_TYPES)[joint.type])
            if joint.limit:
                info += " and has limits [%.3f, %.3f]" % (
                    joint.limit.upper, joint.limit.lower)
            print(info)
        print("The robot's links are:")
        print([link.name for link in self.model.links])



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from compas.utilities import await_callback

from compas_fab.backends.interfaces import AddCollisionMesh
from compas_fab.backends.ros.messages import ApplyPlanningSceneRequest
from compas_fab.backends.ros.messages import ApplyPlanningSceneResponse
from compas_fab.backends.ros.messages import CollisionObject
from compas_fab.backends.ros.messages import PlanningScene
from compas_fab.backends.ros.messages import PlanningSceneWorld
from compas_fab.backends.ros.service_description import ServiceDescription

__all__ = [
    'MoveItAddCollisionMesh',
]


class MoveItAddCollisionMesh(AddCollisionMesh):
    """Callable to add a collision mesh to the planning scene."""
    APPLY_PLANNING_SCENE = ServiceDescription('/apply_planning_scene',
                                              'ApplyPlanningScene',
                                              ApplyPlanningSceneRequest,
                                              ApplyPlanningSceneResponse,
                                              )

    def __init__(self, ros_client):
        self.ros_client = ros_client

    def add_collision_mesh(self, collision_mesh, options=None):
        """Add a collision mesh to the planning scene.

        Parameters
        ----------
        collision_mesh : :class:`compas_fab.robots.CollisionMesh`
            Object containing the collision mesh to be added.
        options : dict, optional
            Unused parameter.

        Returns
        -------
        ``None``
        """
        kwargs = {}
        kwargs['collision_mesh'] = collision_mesh
        kwargs['errback_name'] = 'errback'

        return await_callback(self.add_collision_mesh_async, **kwargs)

    def add_collision_mesh_async(self, callback, errback, collision_mesh):
        co = CollisionObject.from_collision_mesh(collision_mesh)
        co.operation = CollisionObject.ADD
        world = PlanningSceneWorld(collision_objects=[co])
        scene = PlanningScene(world=world, is_diff=True)
        request = dict(scene=scene)
        self.APPLY_PLANNING_SCENE(self.ros_client, request, callback, errback)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from compas_fab.backends.exceptions import BackendError


class PyBulletError(BackendError):
    """Base case for exceptions in ``compas_fab.backends.pybullet``."""
    def __init__(self, message):
        super(PyBulletError, self).__init__(message)


class CollisionError(PyBulletError):
    """Exception raised when two objects have been found to be in collision in PyBullet."""
    def __init__(self, name1, name2):
        message = "Collision between '{}' and '{}'".format(name1, name2)
        super(CollisionError, self).__init__(message)
        self.name1 = name1
        self.name2 = name2


class InverseKinematicsError(PyBulletError):
    """Exception raised when no IK solution can be found in PyBullet."""
    def __init__(self):
        message = "No inverse kinematics solution found."
        super(InverseKinematicsError, self).__init__(message)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from compas.datastructures import Mesh
from compas.geometry import Frame
from compas.geometry import Scale

__all__ = [
    'CollisionMesh',
    'AttachedCollisionMesh',
    'PlanningScene',
]


class CollisionMesh(object):
    """Represents a collision mesh.

    Parameters
    ----------
    mesh : :class:`compas.datastructures.Mesh`
        The collision mesh. Ideally it is as coarse as possible.
    id : :obj:`str`
        The id of the mesh, used to identify it for later operations
        (:meth:`~PlanningScene.add_collision_mesh`,
        :meth:`~PlanningScene.remove_collision_mesh`,
        :meth:`~PlanningScene.append_collision_mesh` etc.)
    frame : :class:`compas.geometry.Frame`, optional
        The frame of the mesh. Defaults to :meth:`~compas.geometry.Frame.worldXY().
    root_name : :obj:`str`
        The name of the root link the collision mesh will be placed in. Defaults
        to `'world'`.

    Attributes
    ----------
    mesh : :class:`compas.datastructures.Mesh`
        The collision mesh. Ideally it is as coarse as possible.
    id : :obj:`str`
        The id of the mesh, used to identify it for later operations
        (:meth:`~PlanningScene.add_collision_mesh`,
        :meth:`~PlanningScene.remove_collision_mesh`,
        :meth:`~PlanningScene.append_collision_mesh` etc.)
    frame : :class:`compas.geometry.Frame`, optional
        The frame of the mesh. Defaults to :meth:`~compas.geometry.Frame.worldXY().
    root_name : :obj:`str`
        The name of the root link the collision mesh will be placed in. Defaults
        to `'world'`.

    Examples
    --------
    >>> mesh = Mesh.from_stl(compas_fab.get('planning_scene/floor.stl'))
    >>> cm = CollisionMesh(mesh, 'floor')
    >>> cm.frame
    Frame(Point(0.000, 0.000, 0.000), Vector(1.000, 0.000, 0.000), Vector(0.000, 1.000, 0.000))
    """

    def __init__(self, mesh, id, frame=None, root_name=None):
        self.id = id
        self.mesh = mesh
        self.frame = frame or Frame.worldXY()
        self.root_name = root_name or 'world'

    def scale(self, scale_factor):
        """Scales the collision mesh uniformly.

        Parameters
        ----------
        scale_factor : :obj:`float`
            Scale factor.
        """
        S = Scale.from_factors([scale_factor] * 3)
        self.mesh.transform(S)

    def scaled(self, scale_factor):
        """Copies the collision mesh, and scales the copy uniformly.

        Parameters
        ----------
        scale_factor : :obj:`float`
            Scale factor.
        """
        self.mesh = self.mesh.copy()
        self.scale(scale_factor)

    def to_data(self):
        """Get the data dictionary that represents the collision mesh.

        This can be used to reconstruct the :class:`CollisionMesh` instance.

        Returns
        -------
        :obj:`dict`
        """
        return self.data

    @classmethod
    def from_data(cls, data):
        """Construct a collision mesh from its data representation.

        Parameters
        ----------
        data : :obj:`dict`
            The data dictionary.

        Returns
        -------
        :class:`CollisionMesh`
             An instance of :class:`CollisionMesh`.
        """
        collision_mesh = cls(None, None)
        collision_mesh.data = data
        return collision_mesh

    @property
    def data(self):
        """:obj:`dict` : The data representing the collision mesh."""
        data_obj = {}
        data_obj['id'] = self.id
        data_obj['mesh'] = self.mesh.to_data()
        data_obj['frame'] = self.frame.to_data()
        data_obj['root_name'] = self.root_name

        return data_obj

    @data.setter
    def data(self, data_obj):
        self.id = data_obj['id']
        self.mesh = Mesh.from_data(data_obj['mesh'])
        self.frame = Frame.from_data(data_obj['frame'])
        self.root_name = data_obj['root_name']


class AttachedCollisionMesh(object):
    """Represents a collision mesh that is attached to a :class:`Robot`'s :class:`~compas.robots.Link`.

    Parameters
    ----------
    collision_mesh : :class:`compas_fab.robots.CollisionMesh`
        The collision mesh to be attached to the robot model.
    link_name : :obj:`str`
        The name of the :class:`~compas.robots.Link` the collision mesh will be
        attached to.
    touch_links : :obj:`list` of :obj:`str`, optional
        The list of link names the collision mesh is allowed to touch. Defaults
        to the link it is attached to.
    weight : :obj:`float`, optional
        The weight of the attached object in kg. Defaults to ``1.0``.

    Attributes
    ----------
    collision_mesh : :class:`compas_fab.robots.CollisionMesh`
        The collision mesh we want to attach.
    link_name : :obj:`str`
        The name of the :class:`~compas.robots.Link` the collision mesh will be
        attached to.
    touch_links : :obj:`list` of :obj:`str`
        The list of link names the collision mesh is allowed to touch. Defaults
        to the link it is attached to.
    weight : :obj:`float`
        The weight of the attached object in kg.

    Examples
    --------
    >>> mesh = Mesh.from_stl(compas_fab.get('planning_scene/cone.stl'))
    >>> cm = CollisionMesh(mesh, 'tip')
    >>> ee_link_name = 'ee_link'
    >>> touch_links = ['wrist_3_link', 'ee_link']
    >>> acm = AttachedCollisionMesh(cm, ee_link_name, touch_links)
    """

    def __init__(self, collision_mesh, link_name, touch_links=None, weight=1.):
        self.collision_mesh = collision_mesh
        if self.collision_mesh:
            self.collision_mesh.root_name = link_name
        self.link_name = link_name
        self.touch_links = touch_links if touch_links else [link_name]
        self.weight = weight

    def to_data(self):
        """Get the data dictionary that represents the attached collision mesh.

        This can be used to reconstruct the :class:`AttachedCollisionMesh` instance.

        Returns
        -------
        :obj:`dict`
        """
        return self.data

    @classmethod
    def from_data(cls, data):
        """Construct an attached collision mesh from its data representation.

        Parameters
        ----------
        data : :obj:`dict`
            The data dictionary.

        Returns
        -------
        :class:`AttachedCollisionMesh`
             An instance of :class:`AttachedCollisionMesh`.
        """
        acm = cls(None, None)
        acm.data = data
        return acm

    @property
    def data(self):
        """:obj:`dict` : The data representing the attached collision mesh."""
        data_obj = {}
        data_obj['collision_mesh'] = self.collision_mesh.to_data()
        data_obj['link_name'] = self.link_name
        data_obj['touch_links'] = self.touch_links
        data_obj['weight'] = self.weight

        return data_obj

    @data.setter
    def data(self, data_obj):
        self.collision_mesh = CollisionMesh.from_data(data_obj['collision_mesh'])
        self.link_name = data_obj['link_name']
        self.touch_links = data_obj['touch_links']
        self.weight = data_obj['weight']


class PlanningScene(object):
    """Represents the planning scene.

    Parameters
    ----------
    robot : :class:`Robot`
        A reference to the robot in the planning scene.

    Attributes
    ----------
    robot : :class:`Robot`
        A reference to the robot in the planning scene.
    client
        A reference to the robot's backend client.
    """

    def __init__(self, robot):
        self.robot = robot

    @property
    def client(self):
        """:class:`compas_fab.backend.RosClient` or :class:`compas_fab.backend.VrepClient` : The backend client."""
        return self.robot.client

    def ensure_client(self):
        """Ensure that the planning scene's robot has a defined client.

        Raises
        ------
        :exc:`Exception`
            If no client is set for planning scene's robot.
        """
        if not self.client:
            raise Exception(
                'This method is only callable once a client is assigned')

    def reset(self):
        """Resets the planning scene, removing all added collision meshes."""
        self.ensure_client()
        self.client.reset_planning_scene()

    def add_collision_mesh(self, collision_mesh, scale=False):
        """Add a collision mesh to the planning scene.

        If there is already a :class:`CollisionMesh` in the
        :class:`PlanningScene` with the same `id` it will be replaced.

        Parameters
        ----------
        collision_mesh : :class:`CollisionMesh`
            The collision mesh we want to add.
        scale : :obj:`bool`, optional
            If ``True``, the mesh will be copied and scaled according to
            the robot's scale factor.

        Returns
        -------
        ``None``

        Examples
        --------
        >>> scene = PlanningScene(robot)
        >>> mesh = Mesh.from_stl(compas_fab.get('planning_scene/floor.stl'))
        >>> cm = CollisionMesh(mesh, 'floor')
        >>> scene.add_collision_mesh(cm)                   # doctest: +SKIP
        """
        self.ensure_client()

        collision_mesh.root_name = self.robot.root_name

        if scale:
            scale_factor = 1. / self.robot.scale_factor
            collision_mesh.scaled(scale_factor)

        self.client.add_collision_mesh(collision_mesh)

    def remove_collision_mesh(self, id):
        """Remove a collision object from the planning scene.

        Parameters
        ----------
        id : :obj:`str`
            The `id` of the :class:`CollisionMesh` instance to remove.

        Returns
        -------
        ``None``

        Examples
        --------
        >>> scene = PlanningScene(robot)
        >>> scene.remove_collision_mesh('floor')           # doctest: +SKIP
        """
        self.ensure_client()
        self.robot.client.remove_collision_mesh(id)

    def append_collision_mesh(self, collision_mesh, scale=False):
        """Append a collision mesh to the planning scene.

        Appends a :class:`CollisionMesh` to the :class:`PlanningScene` using
        `id` as an identifier of a group or cluster of collision meshes. If the group
        does not exist, it will be created implicitly; if it does exist, the meshes will be
        appended to it instead.

        Grouping meshes under a common identifier allows to remove them all
        in one operation, using the :meth:`~PlanningScene.remove_collision_mesh` with
        the group identifier.

        Parameters
        ----------
        collision_mesh : :class:`CollisionMesh`
            The collision mesh we want to append to the :class:`PlanningScene`.
        scale : :obj:`bool`, optional
            If ``True``, the mesh will be copied and scaled according to
            the robot's scale factor.

        Returns
        -------
        ``None``

        Examples
        --------
        >>> scene = PlanningScene(robot)
        >>> mesh = Mesh.from_stl(compas_fab.get('planning_scene/floor.stl'))
        >>> cm = CollisionMesh(mesh, 'floor')
        >>> scene.append_collision_mesh(cm)                # doctest: +SKIP
        """
        self.ensure_client()

        collision_mesh.root_name = self.robot.root_name

        if scale:
            scale_factor = 1. / self.robot.scale_factor
            collision_mesh.scaled(scale_factor)

        self.robot.client.append_collision_mesh(collision_mesh)

    def add_attached_collision_mesh(self, attached_collision_mesh, scale=False):
        """Add an attached collision object to the planning scene.

        Parameters
        ----------
        attached_collision_mesh : :class:`AttachedCollisionMesh`
            The :class:`AttachedCollisionMesh` (a :class:`CollisionMesh`
            attached to a :class:`Robot`'s :class:`~compas.robots.Link`) that
            we want to add to the :class:`PlanningScene`.
        scale : :obj:`bool`, optional
            If ``True``, the mesh will be copied and scaled according to
            the robot's scale factor.

        Returns
        -------
        ``None``

        Examples
        --------
        >>> scene = PlanningScene(robot)
        >>> mesh = Mesh.from_stl(compas_fab.get('planning_scene/cone.stl'))
        >>> cm = CollisionMesh(mesh, 'tip')
        >>> ee_link_name = 'ee_link'
        >>> touch_links = ['wrist_3_link', 'ee_link']
        >>> acm = AttachedCollisionMesh(cm, ee_link_name, touch_links)
        >>> scene.add_attached_collision_mesh(acm)         # doctest: +SKIP
        """
        self.ensure_client()

        if scale:
            scale_factor = 1. / self.robot.scale_factor
            attached_collision_mesh.collision_mesh.scaled(scale_factor)

        self.client.add_attached_collision_mesh(attached_collision_mesh)

    def remove_attached_collision_mesh(self, id):
        """Remove an attached collision object from the planning scene.

        Parameters
        ----------
        id : :obj:`str`
            The `id` of the :class:`CollisionMesh` in the
            :class:`AttachedCollisionMesh` to remove from the
            :class:`PlanningScene`.

        Returns
        -------
        ``None``

        Examples
        --------
        >>> scene = PlanningScene(robot)
        >>> scene.remove_attached_collision_mesh('tip')   # doctest: +SKIP
        """
        self.ensure_client()
        self.client.remove_attached_collision_mesh(id)

    def attach_collision_mesh_to_robot_end_effector(self, collision_mesh, scale=False, group=None):
        """Attaches a collision mesh to the robot's end-effector.

        Parameters
        ----------
        collision_mesh: :class:`CollisionMesh`
            The collision mesh to attach to robot's end effector.
        scale : :obj:`bool`, optional
            If ``True``, the mesh will be copied and scaled according to
            the robot's scale factor.
        group : :obj:`str`
            The planning group with the end effector we want to attach the mesh
            to. Defaults to the robot's main planning group.

        Returns
        -------
        ``None``

        Examples
        --------
        >>> scene = PlanningScene(robot)
        >>> mesh = Mesh.from_stl(compas_fab.get('planning_scene/cone.stl'))
        >>> cm = CollisionMesh(mesh, 'tip')
        >>> group = robot.main_group_name
        >>> scene.attach_collision_mesh_to_robot_end_effector(cm, group=group)

        >>> # check if it's really there
        >>> planning_scene = robot.client.get_planning_scene()
        >>> objects = [c.object['id'] for c in planning_scene.robot_state.attached_collision_objects]
        >>> 'tip' in objects
        True
        >>> scene.remove_attached_collision_mesh('tip')

        >>> # check if it's really gone
        >>> planning_scene = robot.client.get_planning_scene()
        >>> objects = [c.object['id'] for c in planning_scene.robot_state.attached_collision_objects]
        >>> 'tip' in objects
        False
        """
        self.ensure_client()

        if scale:
            scale_factor = 1. / self.robot.scale_factor
            collision_mesh.scaled(scale_factor)

        ee_link_name = self.robot.get_end_effector_link_name(group)
        touch_links = [ee_link_name]
        acm = AttachedCollisionMesh(collision_mesh, ee_link_name, touch_links)
        self.add_attached_collision_mesh(acm)

    def add_attached_tool(self, tool=None):
        """Add the robot's attached tool to the planning scene if tool is set."""
        self.ensure_client()
        if tool:
            self.robot.attach_tool(tool)
        if self.robot.attached_tool:
            for acm in self.robot.attached_tool.attached_collision_meshes:
                self.add_attached_collision_mesh(acm)

    def remove_attached_tool(self):
        """Remove the robot's attached tool from the planning scene."""
        self.ensure_client()
        if self.robot.attached_tool:
            for acm in self.robot.attached_tool.attached_collision_meshes:
                self.remove_attached_collision_mesh(acm.collision_mesh.id)
        self.robot.detach_tool()



"""
Create a position and an orientation constraint from a plane calculated for the group's end-effector link.

COMPAS FAB v0.18.2
"""
import math

from ghpythonlib.componentbase import executingcomponent as component

from compas_fab.ghpython.components import coerce_frame


class ConstraintsFromPlane(component):
    def RunScript(self, robot, plane, group, tolerance_position, tolerance_xaxis, tolerance_yaxis, tolerance_zaxis):
        goal_constraints = None
        if robot and plane:
            tolerance_position = tolerance_position or 0.001
            tolerance_xaxis = tolerance_xaxis or 1.
            tolerance_yaxis = tolerance_yaxis or 1.
            tolerance_zaxis = tolerance_zaxis or 1.

            frame = coerce_frame(plane)
            tolerances_axes = [math.radians(tolerance_xaxis), math.radians(tolerance_yaxis), math.radians(tolerance_zaxis)]
            goal_constraints = robot.constraints_from_frame(frame, tolerance_position, tolerances_axes, group)

        return goal_constraints



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from compas_fab.backends.interfaces import AddCollisionMesh

__all__ = [
    'PyBulletAddCollisionMesh',
]

from compas_fab.backends.pybullet.const import STATIC_MASS


class PyBulletAddCollisionMesh(AddCollisionMesh):
    """Callable to add a collision mesh to the planning scene."""
    def __init__(self, client):
        self.client = client

    def add_collision_mesh(self, collision_mesh, options=None):
        """Add a collision mesh to the planning scene.

        Parameters
        ----------
        collision_mesh : :class:`compas_fab.robots.CollisionMesh`
            Object containing the collision mesh to be added.
        options : dict
            Dictionary containing the following key-value pairs:

            - ``"mass"``: (:obj:`float`) The mass of the object, in kg.
              If `0` is given, (the default), the object added is static.

        Returns
        -------
        ``None``
        """
        options = options or {}
        mesh = collision_mesh.mesh
        name = collision_mesh.id
        frame = collision_mesh.frame
        mass = options.get('mass', STATIC_MASS)

        # mimic ROS' behavior: collision object with same name is replaced
        if name in self.client.collision_objects:
            self.client.remove_collision_mesh(name)

        body_id = self.client.convert_mesh_to_body(mesh, frame, name, mass)
        self.client.collision_objects[name] = [body_id]



"""
********************************************************************************
compas_fab.ghpython
********************************************************************************

.. currentmodule:: compas_fab.ghpython

Artists
-------

In **COMPAS**, the `artists` are classes that assist with the visualization of
datastructures and models, in a way that maintains the data separated from the
specific CAD interfaces, while providing a way to leverage native performance
of the CAD environment.

This package only provides a deprecated alias to the
:class:`~compas_ghpython.artists.RobotModelArtist` implemented in COMPAS.


"""
import compas

if compas.RHINO:
    from .artists import RobotArtist
    from .artists import RobotModelArtist

    from .path_planning import *          # noqa: F401,F403

    __all__ = ['RobotArtist', 'RobotModelArtist']



"""
********************************************************************************
compas_fab.robots
********************************************************************************

.. currentmodule:: compas_fab.robots

This package contains classes for robot modeling and they are used by the
simulation, planning and execution backends to exchange information.

Main classes
------------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    Robot
    RobotSemantics
    Tool
    Duration
    Wrench
    Inertia

Path planning
-------------
.. autosummary::
    :toctree: generated/
    :nosignatures:

    JointTrajectory
    JointTrajectoryPoint
    PathPlan

Planning scene
--------------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    AttachedCollisionMesh
    CollisionMesh
    PlanningScene

Constraints
-----------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    BoundingVolume
    Constraint
    JointConstraint
    OrientationConstraint
    PositionConstraint

Unit conversion
---------------

The unit systems most commonly used in **COMPAS FAB** are **meters** and **radians**.
The following functions help with converting units from one system to the other.

.. autosummary::
    :toctree: generated/
    :nosignatures:

    to_degrees
    to_radians

"""

from .configuration import *          # noqa: F401,F403
from .constraints import *            # noqa: F401,F403
from .path_plan import *              # noqa: F401,F403
from .planning_scene import *         # noqa: F401,F403
from .units import *                  # noqa: F401,F403
from .robot import *                  # noqa: F401,F403
from .semantics import *              # noqa: F401,F403
from .time_ import *                  # noqa: F401,F403
from .tool import *                   # noqa: F401,F403
from .trajectory import *             # noqa: F401,F403
from .wrench import *                 # noqa: F401,F403
from .inertia import *                # noqa: F401,F403

__all__ = [name for name in dir() if not name.startswith('_')]



import csv
import json
import pickle

__all__ = [
    'read_csv_to_dictionary',
    'write_data_to_json',
    'read_data_from_json',
    'write_data_to_pickle',
    'read_data_from_pickle'
]


def read_csv_to_dictionary(csvfile, delimiter=';'):
    """Reads a csv file and returns a dictionary with the respective keys
    specified in the first row of the csv file.

    Parameters
    ----------
    csvfile : str
        The path to csv file.
    delimiter : str, optional
        The character used to separate the values. Default ``;``

    Returns
    -------
    dict
    """
    data = []
    with open(csvfile, mode='r') as infile:
        reader = csv.reader(infile, delimiter=delimiter)
        for rows in reader:
            data.append(rows)
        infile.close()
    data = zip(*data)  # transpose data
    data_dict = {}
    for col in data:
        key = col[0]
        values = list(col[1:])
        data_dict.update({key: values})
    return data_dict


def write_data_to_json(data, file):
    """Write data to json file.

    Parameters
    ----------
    data : dict, list, tuple, str, unicode, int, long, float, boolean, None
        The data to write to json file. Data must be JSON serialisable.
    file : str
        The path where to save the data.
    """
    with open(file, 'w') as f:
        json.dump(data, f)


def read_data_from_json(file):
    """Read data from json file.

    Parameters
    ----------
    file : str
        The path to the json file.

    Returns
    -------
    object
        An object containing the deserialised data from the json file.
    """
    with open(file) as f:
        data = json.load(f)
    return data


def write_data_to_pickle(data, file):
    """Write data to pickle file.

    Parameters
    ----------
    data : object
        The data to write to a pickle file.
    file : str
        The path where to save the data.
    """
    with open(file, 'wb') as f:
        pickle.dump(data, f)


def read_data_from_pickle(file):
    """Read data from pickle file.

    Parameters
    ----------
    file : str
        The path to the pickle file.

    Returns
    -------
    object
        An object containing the reconstituted object hierarchy.
    """
    with open(file, 'rb') as f:
        data = pickle.load(f)
    return data



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


def forward_docstring(backend_feature):
    def dec(obj):
        obj.__doc__ = backend_feature.__dict__[obj.__name__].__doc__
        return obj
    return dec


class ClientInterface(object):
    """Interface for all backend clients.  Forwards all planning services and
    planning scene management to the planner.
    """
    def __init__(self):
        self.planner = PlannerInterface(self)
        # self.control = ControlInterface()

    # ==========================================================================
    # planning services
    # ==========================================================================

    def inverse_kinematics(self, *args, **kwargs):
        """Forwards call to appropriate method in the planner."""
        return self.planner.inverse_kinematics(*args, **kwargs)

    def forward_kinematics(self, *args, **kwargs):
        """Forwards call to appropriate method in the planner."""
        return self.planner.forward_kinematics(*args, **kwargs)

    def plan_cartesian_motion(self, *args, **kwargs):
        """Forwards call to appropriate method in the planner."""
        return self.planner.plan_cartesian_motion(*args, **kwargs)

    def plan_motion(self, *args, **kwargs):
        """Forwards call to appropriate method in the planner."""
        return self.planner.plan_motion(*args, **kwargs)

    # ==========================================================================
    # collision objects and planning scene
    # ==========================================================================

    def get_planning_scene(self, *args, **kwargs):
        """Forwards call to appropriate method in the planner."""
        return self.planner.get_planning_scene(*args, **kwargs)

    def reset_planning_scene(self, *args, **kwargs):
        """Forwards call to appropriate method in the planner."""
        return self.planner.reset_planning_scene(*args, **kwargs)

    def add_collision_mesh(self, *args, **kwargs):
        """Forwards call to appropriate method in the planner."""
        return self.planner.add_collision_mesh(*args, **kwargs)

    def remove_collision_mesh(self, *args, **kwargs):
        """Forwards call to appropriate method in the planner."""
        return self.planner.remove_collision_mesh(*args, **kwargs)

    def append_collision_mesh(self, *args, **kwargs):
        """Forwards call to appropriate method in the planner."""
        return self.planner.append_collision_mesh(*args, **kwargs)

    def add_attached_collision_mesh(self, *args, **kwargs):
        """Forwards call to appropriate method in the planner."""
        return self.planner.add_attached_collision_mesh(*args, **kwargs)

    def remove_attached_collision_mesh(self, *args, **kwargs):
        """Forwards call to appropriate method in the planner."""
        return self.planner.remove_attached_collision_mesh(*args, **kwargs)

#     # ==========================================================================
#     # executing
#     # ==========================================================================
#
#     def execute_joint_trajectory(self, *args, **kwargs):
#         self.control.execute_joint_trajectory(*args, **kwargs)
#
#     def follow_joint_trajectory(self, *args, **kwargs):
#         self.control.follow_joint_trajectory(*args, **kwargs)
#
#     def get_configuration(self, *args, **kwargs):
#         self.control.get_configuration(*args, **kwargs)
#
#
# class ControlInterface(object):
#     def execute_joint_trajectory(self, *args, **kwargs):
#         raise NotImplementedError('Assigned control does not have this feature.')
#
#     def follow_joint_trajectory(self, *args, **kwargs):
#         raise NotImplementedError('Assigned control does not have this feature.')
#
#     def get_configuration(self, *args, **kwargs):
#         raise NotImplementedError('Assigned control does not have this feature.')


class PlannerInterface(object):
    """Interface for all planners associated with a backend client.  Provides default
    behavior for all planning services and planning scene management methods.  To be
    use in conjunction with backend feature interfaces.
    """
    def __init__(self, client):
        super(PlannerInterface, self).__init__()
        self.client = client

    # ==========================================================================
    # planning services
    # ==========================================================================

    def inverse_kinematics(self, *args, **kwargs):
        """Default method for planner.

        Raises
        ------
        Exception
            Planner does not have this feature.
        """
        raise Exception('Assigned planner does not have this feature.')

    def forward_kinematics(self, *args, **kwargs):
        """Default method for planner.

        Raises
        ------
        Exception
            Planner does not have this feature.
        """
        raise Exception('Assigned planner does not have this feature.')

    def plan_motion(self, *args, **kwargs):
        """Default method for planner.

        Raises
        ------
        Exception
            Planner does not have this feature.
        """
        raise Exception('Assigned planner does not have this feature.')

    def plan_cartesian_motion(self, *args, **kwargs):
        """Default method for planner.

        Raises
        ------
        Exception
            Planner does not have this feature.
        """
        raise Exception('Assigned planner does not have this feature.')

    # ==========================================================================
    # collision objects and planning scene
    # ==========================================================================

    def get_planning_scene(self, *args, **kwargs):
        """Default method for planner.

        Raises
        ------
        Exception
            Planner does not have this feature.
        """
        raise Exception('Assigned planner does not have this feature.')

    def reset_planning_scene(self, *args, **kwargs):
        """Default method for planner.

        Raises
        ------
        Exception
            Planner does not have this feature.
        """
        raise Exception('Assigned planner does not have this feature.')

    def add_collision_mesh(self, *args, **kwargs):
        """Default method for planner.

        Raises
        ------
        Exception
            Planner does not have this feature.
        """
        raise Exception('Assigned planner does not have this feature.')

    def remove_collision_mesh(self, *args, **kwargs):
        """Default method for planner.

        Raises
        ------
        Exception
            Planner does not have this feature.
        """
        raise Exception('Assigned planner does not have this feature.')

    def append_collision_mesh(self, *args, **kwargs):
        """Default method for planner.

        Raises
        ------
        Exception
            Planner does not have this feature.
        """
        raise Exception('Assigned planner does not have this feature.')

    def add_attached_collision_mesh(self, *args, **kwargs):
        """Default method for planner.

        Raises
        ------
        Exception
            Planner does not have this feature.
        """
        raise Exception('Assigned planner does not have this feature.')

    def remove_attached_collision_mesh(self, *args, **kwargs):
        """Default method for planner.

        Raises
        ------
        Exception
            Planner does not have this feature.
        """
        raise Exception('Assigned planner does not have this feature.')



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from compas_fab.backends.interfaces.client import forward_docstring
from compas_fab.backends.interfaces.client import PlannerInterface
from compas_fab.backends.vrep.backend_features.vrep_add_attached_collision_mesh import VrepAddAttachedCollisionMesh
from compas_fab.backends.vrep.backend_features.vrep_add_collision_mesh import VrepAddCollisionMesh
from compas_fab.backends.vrep.backend_features.vrep_forward_kinematics import VrepForwardKinematics
from compas_fab.backends.vrep.backend_features.vrep_inverse_kinematics import VrepInverseKinematics
from compas_fab.backends.vrep.backend_features.vrep_plan_motion import VrepPlanMotion
from compas_fab.backends.vrep.backend_features.vrep_remove_collision_mesh import VrepRemoveCollisionMesh


class VrepPlanner(PlannerInterface):
    """Implement the planner backend interface for V-REP
    """
    def __init__(self, client):
        super(VrepPlanner, self).__init__(client)

    @forward_docstring(VrepForwardKinematics)
    def forward_kinematics(self, *args, **kwargs):
        return VrepForwardKinematics(self.client)(*args, **kwargs)

    @forward_docstring(VrepInverseKinematics)
    def inverse_kinematics(self, *args, **kwargs):
        return VrepInverseKinematics(self.client)(*args, **kwargs)

    @forward_docstring(VrepPlanMotion)
    def plan_motion(self, *args, **kwargs):
        return VrepPlanMotion(self.client)(*args, **kwargs)

    @forward_docstring(VrepPlanMotion)
    def plan_motion_to_config(self, *args, **kwargs):
        return VrepPlanMotion(self.client).plan_motion_to_config(*args, **kwargs)

    @forward_docstring(VrepAddAttachedCollisionMesh)
    def add_attached_collision_mesh(self, *args, **kwargs):
        return VrepAddAttachedCollisionMesh(self.client)(*args, **kwargs)

    @forward_docstring(VrepAddAttachedCollisionMesh)
    def pick_building_member(self, *args, **kwargs):
        return VrepAddAttachedCollisionMesh(self.client).pick_building_member(*args, **kwargs)

    @forward_docstring(VrepAddCollisionMesh)
    def add_collision_mesh(self, *args, **kwargs):
        return VrepAddCollisionMesh(self.client)(*args, **kwargs)

    @forward_docstring(VrepRemoveCollisionMesh)
    def remove_collision_mesh(self, *args, **kwargs):
        return VrepRemoveCollisionMesh(self.client)(*args, **kwargs)



from __future__ import absolute_import

from compas.geometry import Frame

import compas_fab.robots

from .std_msgs import Header
from .std_msgs import ROSmsg


class Point(ROSmsg):
    """https://docs.ros.org/api/geometry_msgs/html/msg/Point.html
    """

    def __init__(self, x, y, z):
        self.x = x
        self.y = y
        self.z = z

    @classmethod
    def from_msg(cls, msg):
        x, y, z = msg['x'], msg['y'], msg['z']
        return cls(x, y, z)


class Quaternion(ROSmsg):
    """https://docs.ros.org/api/geometry_msgs/html/msg/Quaternion.html
    """

    def __init__(self, x=0., y=0., z=0., w=1.):
        self.x = x
        self.y = y
        self.z = z
        self.w = w

    @classmethod
    def from_frame(cls, frame):
        qw, qx, qy, qz = frame.quaternion
        return cls(qx, qy, qz, qw)


class Pose(ROSmsg):
    """https://docs.ros.org/api/geometry_msgs/html/msg/Pose.html
    """

    def __init__(self, position=None, orientation=None):
        self.position = position if position else Point(0, 0, 0)
        self.orientation = orientation if orientation else Quaternion(0, 0, 0, 1)

    @classmethod
    def from_frame(cls, frame):
        point = frame.point
        qw, qx, qy, qz = frame.quaternion
        return cls(Point(*list(point)), Quaternion(qx, qy, qz, qw))

    @property
    def frame(self):
        point = [self.position.x, self.position.y, self.position.z]
        quaternion = [self.orientation.w, self.orientation.x,
                      self.orientation.y, self.orientation.z]
        return Frame.from_quaternion(quaternion, point=point)

    @classmethod
    def from_msg(cls, msg):
        position = Point.from_msg(msg['position'])
        orientation = Quaternion.from_msg(msg['orientation'])
        return cls(position, orientation)


class PoseStamped(ROSmsg):
    """https://docs.ros.org/api/geometry_msgs/html/msg/PoseStamped.html
    """

    def __init__(self, header=None, pose=None):
        self.header = header or Header()
        self.pose = pose or Pose()

    @classmethod
    def from_msg(cls, msg):
        header = Header.from_msg(msg['header'])
        pose = Pose.from_msg(msg['pose'])
        return cls(header, pose)


class Vector3(ROSmsg):
    """https://docs.ros.org/api/geometry_msgs/html/msg/Vector3.html
    """

    def __init__(self, x=0., y=0., z=0.):
        self.x = x
        self.y = y
        self.z = z

    @classmethod
    def from_msg(cls, msg):
        x, y, z = msg['x'], msg['y'], msg['z']
        return cls(x, y, z)


class Transform(ROSmsg):
    """https://docs.ros.org/api/geometry_msgs/html/msg/Transform.html
    """

    def __init__(self, translation=None, rotation=None):
        self.translation = translation or Vector3()
        self.rotation = rotation or Quaternion()


class Twist(ROSmsg):
    """https://docs.ros.org/api/geometry_msgs/html/msg/Twist.html
    """

    def __init__(self, linear=None, angular=None):
        self.linear = linear or Vector3()
        self.angular = angular or Vector3()


class Wrench(ROSmsg):
    """https://docs.ros.org/api/geometry_msgs/html/msg/Wrench.html

    This represents force in free space, separated into its linear and angular parts.

    Examples
    --------
    >>> wrench = compas_fab.robots.Wrench([0, 0, -98], [0, 0, 0])
    >>> ros_wrench = Wrench.from_wrench(wrench)
    >>> ros_wrench.msg
    {'force': {'x': 0.0, 'y': 0.0, 'z': -98.0}, 'torque': {'x': 0.0, 'y': 0.0, 'z': 0.0}}
    >>> ros_wrench.wrench
    Wrench(Vector(0.000, 0.000, -98.000), Vector(0.000, 0.000, 0.000))
    """

    def __init__(self, force=None, torque=None):
        self.force = force or Vector3()
        self.torque = torque or Vector3()

    @classmethod
    def from_msg(cls, msg):
        force = Vector3.from_msg(msg['force'])
        torque = Vector3.from_msg(msg['torque'])
        return cls(force, torque)

    @classmethod
    def from_wrench(cls, wrench):
        force = wrench.force
        torque = wrench.torque
        return cls(Vector3(*list(force)), Vector3(*list(torque)))

    @property
    def wrench(self):
        force = [self.force.x, self.force.y, self.force.z]
        torque = [self.torque.x, self.torque.y, self.torque.z]
        return compas_fab.robots.Wrench(force, torque)


class WrenchStamped(ROSmsg):
    """https://docs.ros.org/api/geometry_msgs/html/msg/WrenchStamped.html

    A wrench with reference coordinate frame and timestamp.
    """

    def __init__(self, header=None, wrench=None):
        self.header = header or Header()
        self.wrench = wrench or Wrench()

    @classmethod
    def from_msg(cls, msg):
        header = Header.from_msg(msg['header'])
        wrench = Wrench.from_msg(msg['wrench'])
        return cls(header, wrench)


class Inertia(ROSmsg):
    """https://docs.ros.org/api/geometry_msgs/html/msg/Inertia.html

    Examples
    --------
    >>> inertia = compas_fab.robots.Inertia([[0] * 3] * 3, 1., [0.1, 3.1, 4.4])
    >>> ros_inertia = Inertia.from_inertia(inertia)
    >>> ros_inertia.msg
    {'m': 1.0, 'com': {'x': 0.1, 'y': 3.1, 'z': 4.4}, 'ixx': 0.0, 'ixy': 0.0, 'ixz': 0.0, 'iyy': 0.0, 'iyz': 0.0, 'izz': 0.0}
    >>> ros_inertia.inertia
    Inertia([[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]], 1.0, Point(0.100, 3.100, 4.400))
    """

    def __init__(self, m=0.0, com=None, ixx=0., ixy=0., ixz=0., iyy=0., iyz=0., izz=0.):
        self.m = float(m)  # Mass [kg]
        self.com = com or Vector3()  # Center of mass [m]
        self.ixx = float(ixx)
        self.ixy = float(ixy)
        self.ixz = float(ixz)
        self.iyy = float(iyy)
        self.iyz = float(iyz)
        self.izz = float(izz)

    @classmethod
    def from_msg(cls, msg):
        com = Vector3.from_msg(msg['com'])
        return cls(msg['m'], com, msg['ixx'], msg['ixy'], msg['ixz'], msg['iyy'], msg['iyz'], msg['izz'])

    @classmethod
    def from_inertia(cls, inertia):
        m = inertia.mass
        com = Vector3(*list(inertia.center_of_mass))
        ixx, ixy, ixz = inertia.inertia_tensor[0]
        ixy, iyy, iyz = inertia.inertia_tensor[1]
        izz = inertia.inertia_tensor[2][2]
        return cls(m, com, ixx, ixy, ixz, iyy, iyz, izz)

    @property
    def inertia(self):
        inertia_tensor = [[self.ixx, self.ixy, self.ixz],
                          [self.ixy, self.iyy, self.iyz],
                          [self.ixz, self.iyz, self.izz]]
        return compas_fab.robots.Inertia(inertia_tensor, self.m, [self.com.x, self.com.y, self.com.z])


if __name__ == "__main__":
    import doctest
    doctest.testmod()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from compas_fab.backends.vrep.helpers import floats_to_vrep
from compas_fab.backends.interfaces import AddCollisionMesh


__all__ = [
    'VrepAddCollisionMesh',
]


class VrepAddCollisionMesh(AddCollisionMesh):
    """Callable to add a mesh to the 3D scene.
    """
    def __init__(self, client):
        self.client = client

    def add_collision_mesh(self, collision_mesh, options=None):
        """Adds meshes to the 3D scene.

        Args:
            collision_mesh (:obj:`list` of :class:`compas.datastructures.Mesh`): List
                of meshes to add to the current simulation scene.
            options (:obj:`dict`): Unused parameter.

        Returns:
            list: List of object handles (identifiers) assigned to the meshes.

        .. note::
            All meshes are automatically removed from the scene when the simulation ends.
        """
        return self.add_meshes(collision_mesh)

    def add_meshes(self, meshes):
        """Adds meshes to the 3D scene.

        Args:
            meshes (:obj:`list` of :class:`compas.datastructures.Mesh`): List
                of meshes to add to the current simulation scene.

        Returns:
            list: List of object handles (identifiers) assigned to the meshes.

        .. note::
            All meshes are automatically removed from the scene when the simulation ends.
        """
        mesh_handles = []

        for mesh in meshes:
            if not mesh.is_trimesh():
                raise ValueError('The V-REP client only supports tri-meshes')

            vertices, faces = mesh.to_vertices_and_faces()
            vrep_packing = (floats_to_vrep([item for sublist in vertices for item in sublist], self.client.scale) +
                            [item for sublist in faces for item in sublist])
            params = [[len(vertices) * 3, len(faces) * 4], vrep_packing]
            handles = self.client.run_child_script('buildMesh',
                                                   params[0],
                                                   params[1],
                                                   [])[1]
            mesh_handles.extend(handles)
            self.client._added_handles.extend(handles)

        return mesh_handles



from __future__ import print_function

import math

import compas
from compas.geometry import Vector
from compas.geometry import cross_vectors

if not compas.IPY:
    from scipy import stats
else:
    stats = None

# TODO: move to some constants file? g is also defined in scipy.constants
# standard acceleration of gravity [m/s**2]
g = 9.80665
gravity_vector = Vector(0, 0, -g)


__all__ = ['Wrench']


class Wrench():
    """A wrench represents force in free space, separated into its linear (force) and angular (torque) parts.

    Attributes
    ----------
    force : :class:`Vector`
        [Fx, Fy, Fz] force vector in Newtons
    torque : :class:`Vector`
        [Tx, Ty, Tz] moments vector in Newton-meters

    Examples
    --------
    >>> from compas.geometry import Vector
    >>> w = Wrench([1, 2, 3], [0.1, 0.2, 0.3])
    >>> w = Wrench(Vector(1, 2, 3), Vector(0.1, 0.2, 0.3))

    """

    def __init__(self, force, torque):
        self.force = force
        self.torque = torque

    # ==========================================================================
    # factory
    # ==========================================================================

    @classmethod
    def from_list(cls, values):
        """Construct a wrench from a list of 6 :obj:`float` values.

        Parameters
        ----------
        values : :obj:`list` of :obj:`float`
            The list of 6 values representing a wrench.

        Returns
        -------
        :class:`Wrench`
            The constructed wrench.

        Examples
        --------
        >>> w = Wrench.from_list([1, 2, 3, 0.1, 0.2, 0.3])
        """
        force = values[0:3]
        torque = values[3:6]
        return cls(force, torque)

    @classmethod
    def from_data(cls, data):
        """Construct a wrench from its data representation.

        Parameters
        ----------
        data : :obj:`dict`
            The data dictionary.

        Returns
        -------
        :class:`Wrench`
            The constructed wrench.

        Examples
        --------
        >>> data = {"force": [1, 2, 3], "torque": [0.1, 0.2, 0.3]}
        >>> w = Wrench.from_data(data)
        """
        force = data["force"]
        torque = data["torque"]
        return cls(force, torque)

    @classmethod
    def by_samples(cls, wrenches, proportion_to_cut=0.1):
        """
        Construct the wrench by sampled data, allowing to filter.

        Parameters
        ----------
        wrenches : list of :class:`Wrench`
            List of wrenches.
        proportion_to_cut : :obj:`float`
            Fraction to cut off of both tails of the distribution

        Returns
        -------
        Wrench
            The mean wrench after trimming distribution from both tails.

        Examples
        --------
        >>> w1 = Wrench([1, 1, 1], [.1,.1,.1])
        >>> w2 = Wrench([2, 2, 2], [.2,.2,.2])
        >>> w3 = Wrench([3, 3, 3], [.3,.3,.3])
        >>> w = Wrench.by_samples([w1, w2, w3])
        >>> print(w)
        Wrench(Vector(2.000, 2.000, 2.000), Vector(0.200, 0.200, 0.200))
        """
        if not stats:
            raise NotImplementedError("Not supported on this platform")

        forces = [w.force for w in wrenches]
        torques = [w.torque for w in wrenches]
        force = stats.trim_mean(forces, proportion_to_cut, axis=0).tolist()
        torque = stats.trim_mean(torques, proportion_to_cut, axis=0).tolist()
        return cls(force, torque)

    # ==========================================================================
    # descriptors
    # ==========================================================================

    @property
    def force(self):
        return self._force

    @force.setter
    def force(self, vector):
        force = Vector(*list(vector))
        self._force = force

    @property
    def torque(self):
        return self._torque

    @torque.setter
    def torque(self, vector):
        torque = Vector(*list(vector))
        self._torque = torque

    @property
    def data(self):
        """Returns the data dictionary that represents the wrench.

        Returns
        -------
        dict
            The wrench data.
        """
        return {'force': list(self.force),
                'torque': list(self.torque)}

    def to_data(self):
        """Returns the data dictionary that represents the wrench.

        Returns
        -------
        dict
            The wrench data.
        """
        return self.data

    # ==========================================================================
    # access
    # ==========================================================================

    def __iter__(self):
        return iter(list(self.force) + list(self.torque))

    def __getitem__(self, item):
        w = list(self)
        return w[item]

    # ==========================================================================
    # representation
    # ==========================================================================

    def __repr__(self):
        return "Wrench({0}, {1})".format(self.force, self.torque)

    # ==========================================================================
    # helpers
    # ==========================================================================

    def copy(self):
        """Make a copy of this ``Wrench``.

        Returns
        -------
        Wrench
            The copy.
        """
        cls = type(self)
        return cls(self.force.copy(), self.torque.copy())

    # ==========================================================================
    # operators
    # ==========================================================================

    def __mul__(self, n):
        """Create a ``Wrench`` multiplied by the given factor.

        Parameters
        ----------
        n : float
            The multiplication factor.

        Returns
        -------
        Wrench
            The resulting new ``Wrench``.

        """
        return Wrench(self.force * n, self.torque * n)

    def __add__(self, other):
        """Return a ``Wrench`` that is the the sum of this ``Wrench`` and another wrench.

        Parameters
        ----------
        other : wrench
            The wrench to add.

        Returns
        -------
        Wrench
            The resulting new ``Wrench``.

        """
        return Wrench(
            self.force + other.force,
            self.torque + other.torque)

    def __sub__(self, other):
        """Return a ``Wrench`` that is the the difference between this ``Wrench`` and another wrench.

        Parameters
        ----------
        other : wrench
            The wrench to subtract.

        Returns
        -------
        Wrench
            The resulting new ``Wrench``.
        """
        return Wrench(
            self.force - other.force,
            self.torque - other.torque)

    def __neg__(self):
        """Return the negated ``Wrench``.

        Returns
        -------
        Wrench
            The negated ``Wrench``.
        """
        return Wrench(
            self.force * -1.0,
            self.torque * -1.0)

    def __len__(self):
        return 6

    # ==========================================================================
    # comparison
    # ==========================================================================

    def __eq__(self, other, tol=1e-05):
        for a, b in zip(list(self), list(other)):
            if math.fabs(a - b) > tol:
                return False
        return True

    def __ne__(self, other, tol=1e-05):
        return not self.__eq__(other, tol)

    # ==========================================================================
    # transformations
    # ==========================================================================

    def transform(self, transformation):
        """Transforms a `Wrench` with the transformation.

        Parameters
        ----------
        transformation : :class:`Transformation`
            The transformation to transform the `Wrench`.

        Returns
        -------
        None

        Examples
        --------
        >>> R = Rotation.from_axis_and_angle([1, 0, 0], math.pi)
        >>> w = Wrench([1, 2, 3], [0.1, 0.2, 0.3])
        >>> w.transform(R)
        """
        self.force.transform(transformation)
        self.torque.transform(transformation)

    def transformed(self, transformation):
        """Returns a transformed copy of the `Wrench`.

        Parameters
        ----------
        transformation : :class:`Transformation`
            The transformation to transform the `Wrench`.

        Returns
        -------
        Wrench
            The transformed wrench.

        Examples
        --------
        >>> R = Rotation.from_axis_and_angle([1, 0, 0], math.pi)
        >>> w1 = Wrench([1, 2, 3], [0.1, 0.2, 0.3])
        >>> w2 = w1.transformed(R)
        """
        wrench = self.copy()
        wrench.transform(transformation)
        return wrench

    def gravity_compensated(self, ft_sensor_frame, mass, center_of_mass):
        """Removes the force and torque in effect of gravity from the wrench.

        Parameters
        ----------
        ft_sensor_frame : :class:`compas.geometry.Frame`
            The coordinate frame of the force torque sensor.
        mass: float
            The mass of the object in kg.
        center_of_mass : :class:`Point`
            The center of mass of the object in meters.

        Returns
        -------
        Wrench
            The gravity compensated wrench.

        Examples
        --------
        >>> mass, com = 10, [0, 0, 1]
        >>> f = Frame([0, 0, 0], [1, 0, 0], [0, 1, 0])
        >>> w = Wrench([0, 0, -98], [0, 0, 0])
        >>> w.gravity_compensated(f, mass, com)
        Wrench(Vector(0.000, 0.000, 0.066), Vector(0.000, 0.000, 0.000))

        >>> mass, com = 10, [1, 1, 1]
        >>> f = Frame([0, 0, 0], [1, 0, 0], [0, 1, 0])
        >>> w = Wrench([0, 0, -98], [-98, 98, 0])
        >>> w.gravity_compensated(f, mass, com)
        Wrench(Vector(0.000, 0.000, 0.066), Vector(-88.193, 88.193, 0.000))

        Notes
        -----
        For more info, see [1]_.

        References
        ----------
        .. [1] Vougioukas S., *Bias Estimation and Gravity Compensation For
            Force-Torque Sensors*,
            Available at: https://citeseerx.ist.psu.edu/viewdoc/summary?doi=10.1.1.552.109.
        """
        # transform gravity vector to FT Sensor coordinate system (FTSCS)
        gravity_vector_FTSCS = ft_sensor_frame.to_local_coordinates(gravity_vector)

        # F gravity compensation, F = gravity * mass
        force_gravity = gravity_vector_FTSCS * mass

        # torque gravity compensation, T = (lever_arm * mass) X gravity_vector_FTSCS
        torque_gravity = Vector(*cross_vectors((center_of_mass * mass), gravity_vector_FTSCS))

        return Wrench(self.force - force_gravity, self.torque - torque_gravity)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from compas.geometry import Frame


__all__ = ['pose_from_frame', 'frame_from_pose']


def pose_from_frame(frame):
    """Returns a PyBullet pose from a frame.

    Parameters
    ----------
    frame : :class:`compas.geometry.Frame`

    Returns
    -------
    point, quaternion : tuple
    """
    return list(frame.point), frame.quaternion.xyzw


def frame_from_pose(pose):
    """Returns a frame from a PyBullet pose.

    Parameters
    ----------
    pose : tuple

    Returns
    -------
    :class:`compas.geometry.Frame`
    """
    point, (x, y, z, w) = pose
    return Frame.from_quaternion([w, x, y, z], point=point)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .std_msgs import ROSmsg
from .std_msgs import Header

from .geometry_msgs import Pose


class Octomap(ROSmsg):
    """https://docs.ros.org/kinetic/api/octomap_msgs/html/msg/Octomap.html"""

    def __init__(self, header=None, binary=False, id='', resolution=0., data=None):
        self.header = header or Header()  # Header
        self.binary = binary  # Flag to denote a binary (only free/occupied) or full occupancy octree (.bt/.ot file)
        self.id = id  # Class id of the contained octree
        self.resolution = resolution  # Resolution (in m) of the smallest octree nodes
        self.data = data or []  # binary serialization of octree, use conversions.h to read and write octrees


class OctomapWithPose(ROSmsg):
    """https://docs.ros.org/kinetic/api/octomap_msgs/html/msg/OctomapWithPose.html"""

    def __init__(self, header=None, origin=None, octomap=None):
        self.header = header or Header()  # Header
        self.origin = origin or Pose()  # geometry_msgs/Pose  The pose of the octree with respect to the header frame
        self.octomap = octomap or Octomap()  # octomap_msgs/Octomap  The actual octree msg



"""
Add or remove a collision mesh from the planning scene.

COMPAS FAB v0.18.2
"""
from compas_rhino.geometry import RhinoMesh
from ghpythonlib.componentbase import executingcomponent as component

from compas_fab.robots import CollisionMesh


class CollisionMeshComponent(component):
    def RunScript(self, scene, M, name, add, remove):
        ok = False
        if scene and M and name:
            mesh = RhinoMesh.from_geometry(M).to_compas()
            collision_mesh = CollisionMesh(mesh, name)
            if add:
                scene.add_collision_mesh(collision_mesh)
                ok = True
            if remove:
                scene.remove_collision_mesh(name)
                ok = True
        return ok
